# -*- coding: utf8 -*-
import datetime
import logging
from uuid import uuid4
import six
from .base_metadata_db import BaseMetadataDB, MetadataTypeNotSupported, MetadataFieldNotFound, MetadataOperationError
from ..bigquery_mixin import BigQueryMixin, BqJob
import re


# noinspection SqlNoDataSourceInspection
class BigQueryMetadataDB(BaseMetadataDB, BigQueryMixin):
    STAGING_TABLE_NAME = 'staging'
    STAGING_COMMIT = 'staging'
    METADATA_TABLE_NAME = 'metadata'
    INDEX_TABLE_NAME = 'index'
    STAGING_INDEX_TABLE_NAME = 'staging_index'
    TEMP_TABLE_TTL = 10 * 60

    def __init__(self, connection, version=None, version_ts_lookup=None, delete_temp_on_commit=True, isolated=False):
        self.__version = version
        self.__staging_table = None
        self.__metadata_table = None
        self.__prev_table_info = None
        self.__delete_temp_on_commit = delete_temp_on_commit
        self.__isolated = isolated
        super(BigQueryMetadataDB, self).__init__(connection, version_ts_lookup)

    def __get_staging_table_name(self, version=None):
        version = version or self.__version or 0
        return '%s_%s' % (self.STAGING_TABLE_NAME, version)

    def __get_staging_index_table_name(self):
        version = self.__version or 0
        return '%s_%s' % (self.STAGING_INDEX_TABLE_NAME, version)

    @classmethod
    def _default_table_schema(cls):
        from google.cloud import bigquery

        schema = (
            bigquery.SchemaField('_sha', 'STRING', 'REQUIRED'),
            bigquery.SchemaField('_commit_sha', 'STRING', 'REQUIRED'),
            bigquery.SchemaField('_ts', 'TIMESTAMP', 'REQUIRED'),
            bigquery.SchemaField('_hash', 'STRING', 'REQUIRED'),
        )

        return schema

    def _create_table(self):
        schema = self._default_table_schema()

        self.__metadata_table = self._create_specific_table(self.METADATA_TABLE_NAME, schema)
        self.__staging_table = self._create_specific_table(self.__get_staging_table_name(), self.__metadata_table.schema)

    def __get_staging_table(self):
        self.__staging_table = self._get_specific_table(self.__get_staging_table_name())
        return self.__staging_table

    def __using_staging_table(self):
        if self.__staging_table is None:
            self.__get_staging_table()

    def __get_metadata_table(self):
        self.__metadata_table = self._get_specific_table(self.METADATA_TABLE_NAME)
        return self.__metadata_table

    def __using_metadata_table(self):
        if self.__metadata_table is None:
            self.__get_metadata_table()

    @classmethod
    def __value_to_sql_type(cls, column_value):
        if isinstance(column_value, six.string_types):
            return 'STRING'

        if isinstance(column_value, six.integer_types):
            return 'INTEGER'

        if isinstance(column_value, float):
            return 'FLOAT'

        raise MetadataTypeNotSupported('UNKNOWN TYPE %s' % type(column_value))

    def _patch_table_with_new_schema(self, table, schema):
        bq_client = self._connection

        table.schema = schema
        bq_client.update_table(table, ['schema'])

    @classmethod
    def __validate_new_columns(cls, table, new_columns):
        schema = table.schema[:]
        existing_fields = {field.name: field for field in schema}
        actual_new_columns = []
        for new_column in new_columns:
            if new_column.name in existing_fields:
                existing_field = existing_fields[new_column.name]
                if new_column.field_type.lower() != existing_field.field_type.lower():
                    raise MetadataOperationError('field %s changed type from %s to %s' % (
                        new_column.name, new_column.field_type, existing_field.field_type))

                continue

            actual_new_columns.append(new_column)

        return actual_new_columns

    def __add_columns(self, table, new_columns, refresh_table_method):
        from google.api_core.exceptions import PreconditionFailed

        logging.info('add columns %s', new_columns)

        if len(new_columns) == 0:
            return

        retry = 3

        while retry > 0:
            actual_new_columns = self.__validate_new_columns(table, new_columns)

            if len(actual_new_columns) == 0:
                return

            schema = table.schema[:]
            actual_new_columns = sorted(actual_new_columns, key=lambda field: field.name)
            schema.extend(actual_new_columns)
            try:
                self._patch_table_with_new_schema(table, schema)
                break
            except PreconditionFailed:
                logging.info('table %s precondition failed, retrying', self.__staging_table.table_id)
                table = refresh_table_method()
                retry -= 1
                continue

    def __columns_from_data_object(self, data_object):
        from google.cloud import bigquery

        columns = []
        for column_name, column_value in data_object.items():
            column_name = self.common_name_to_bq_field_name(column_name)
            columns.append(bigquery.SchemaField(column_name, self.__value_to_sql_type(column_value)))

        return columns

    def _add_missing_columns(self, data_object):
        self.__using_staging_table()

        columns = self.__columns_from_data_object(data_object)
        self.__add_columns(self.__staging_table, columns, self.__get_staging_table)

    @classmethod
    def _fields_for_select(cls, fields):
        return ','.join(['`%s`' % field for field in fields])

    @classmethod
    def _schema_fields(cls, schema):
        return [field.name for field in schema]

    def _merge_rows(self, src_table, schema=None):
        from google.api_core.exceptions import BadRequest

        self.__using_staging_table()
        self.__using_metadata_table()

        built_in_fields = set(self._schema_fields(self._default_table_schema()))

        fields_new = set(self._schema_fields(schema or src_table.schema))
        staging_fields = set(self._schema_fields(self.__staging_table.schema)) - built_in_fields
        metadata_fields = set(self._schema_fields(self.__metadata_table.schema)) - built_in_fields

        only_in_staging_fields = staging_fields - fields_new
        only_in_metadata_fields = metadata_fields - only_in_staging_fields - fields_new

        all_meta_fields = []
        all_meta_fields.extend(fields_new)
        all_meta_fields.extend(staging_fields)
        all_meta_fields.extend(metadata_fields)

        all_meta_fields = sorted(set(all_meta_fields))

        select_fields = []
        for field in all_meta_fields:
            select_field = 'TO_JSON_STRING(`{field}`)'.format(field=field)
            select_fields.append(select_field)

        hash_select = 'TO_HEX(SHA1(CONCAT({fields}))) as _hash'.format(fields=','.join(select_fields))

        with self._connection.get_cursor() as bq_dataset:
            src_query = """
                #standardSQL
                #_merge_rows
                SELECT *, {hash_select}, CURRENT_TIMESTAMP() as _ts, 'staging' as _commit_sha
                FROM (
                    SELECT *  
                    FROM `{dataset_name}.{src_table}`
                """

            if len(only_in_staging_fields) > 0:
                src_query += """
                    LEFT JOIN (
                        SELECT `_sha`, {only_in_staging_fields}
                        FROM `{dataset_name}.{metadata_staging_index_table_name}`
                    )
                    USING (_sha)
                """

            if len(only_in_metadata_fields) > 0:
                src_query += """
                    LEFT JOIN (
                        SELECT `_sha`, {only_in_metadata_fields}
                        FROM `{dataset_name}.{metadata_table_name}`
                    )
                    USING (_sha)                    
                """

            src_query += ')'

            src_query = src_query.format(
                hash_select=hash_select,
                dataset_name=bq_dataset.dataset_id,
                src_table_columns=self._fields_for_select(fields_new),
                only_in_staging_fields=self._fields_for_select(only_in_staging_fields),
                only_in_metadata_fields=self._fields_for_select(only_in_metadata_fields),
                src_table=src_table.table_id,
                **self.__get_table_names()
            )

            new_staging_fields = fields_new - staging_fields - built_in_fields
            new_staging_fields = list(self._get_fields_schema(new_staging_fields, src_table))
            self.__add_columns(self.__staging_table, new_staging_fields, self.__get_staging_table)

            job = self._async_copy_table_data(src_query, (), self._get_table_ref(self.__get_staging_table_name()))

            try:
                BqJob(job).wait()
            except BadRequest as ex:
                logging.info(str(ex))
                message = ex.errors[0]['message'] if len(ex.errors) > 0 else 'Unknown error'
                raise MetadataOperationError(message)

    @classmethod
    def _delete_temp_table(cls, project, table_ref):
        from controllers.legit import BigQueryConnection
        from google.cloud import bigquery

        bq_client = bigquery.Client(project=project)

        BigQueryConnection.delete_table_using_connection(bq_client, table_ref)

    def _get_temp_table(self, prefix):
        prefix = 'tmp_' + prefix
        from deferred import do_defer

        temp_table_ref = self._get_table_ref(uuid4().hex, prefix=prefix)

        do_defer(
            self._delete_temp_table, temp_table_ref.project, temp_table_ref,
            _countdown=self.TEMP_TABLE_TTL, _tag=temp_table_ref.project)

        return temp_table_ref

    def add_data_using_url(self, metadata_url, dry_mode=False):
        temp_table_ref = self._get_temp_table(prefix='add_data')

        job = self._async_load_job(metadata_url, table_ref=temp_table_ref)

        result = BqJob(job).wait()

        logging.info('result %s', result)

        if not dry_mode:
            bq_client = self._connection
            dest_table = bq_client.get_table(result.destination)
            self._merge_rows(dest_table)

    def _add_data(self, flatten_data_list):
        rows = []

        now = datetime.datetime.utcnow()

        for flatten_data_list in flatten_data_list:
            row = []
            for field in self.__staging_table.schema:
                common_field_name = self.bq_field_name_to_common_name(field.name)
                if common_field_name == '_ts':
                    row.append(now)
                else:
                    row.append(flatten_data_list.get(common_field_name))

            rows.append(row)

        bq_client = self._connection
        bq_client.create_rows(self.__staging_table, rows)

    def __truncate_staging(self):
        import google.cloud.exceptions

        if not self.__delete_temp_on_commit:
            logging.debug('meta: delete_temp_on_commit')
            return

        logging.info('truncate metadata staging')
        bq_client = self._connection
        # noinspection PyUnresolvedReferences
        try:
            bq_client.delete_table(self._get_table_ref(self.__get_staging_table_name()))
        except google.cloud.exceptions.NotFound:
            logging.info('table %s not found', self.__get_staging_table_name())
            pass

    def end_commit(self):
        logging.debug('bq end commit meta')

        if not self.__isolated:
            self.__truncate_staging()

    def begin_commit(self, commit_sha, tree_id, ts):
        logging.debug('bq begin commit meta %s %s (%s, %s)', commit_sha, tree_id, self.__isolated, self.__version)

        from google.cloud import bigquery

        self.__using_staging_table()
        self.__using_metadata_table()

        if self.__staging_table.schema != self.__metadata_table.schema:
            self.__add_columns(self.__metadata_table, self.__staging_table.schema, self.__get_metadata_table)

        with self._connection.get_cursor() as bq_dataset:
            if not self.__isolated:
                # create the table for the next staging
                self._create_specific_table(self.__get_staging_table_name(version=self.__version + 1), self.__staging_table.schema)

            src_query = """
                #standardSQL
                #begin_commit
                SELECT @commit_sha as _commit_sha, @ts as _ts, TableA.*
                # Find the latest metadata given in the staging meta
                FROM (
                  SELECT * EXCEPT(row_number)
                  FROM (
                    # Find the latest metadata given in the staging meta
                    SELECT * EXCEPT(_commit_sha, _ts) , ROW_NUMBER() OVER (PARTITION BY _sha ORDER BY _ts DESC) row_number
                    FROM {dataset_name}.{metadata_staging_index_table_name}
                  )
                  WHERE row_number = 1
                ) TableA
                LEFT OUTER JOIN (
                  SELECT * EXCEPT(row_number)
                  FROM (
                    SELECT *, ROW_NUMBER() OVER (PARTITION BY _sha ORDER BY _ts DESC) row_number
                    FROM {dataset_name}.{metadata_table_name}
                  )
                  WHERE row_number = 1
                ) TableB
                USING (_sha, _hash)
                WHERE TableB._sha IS NULL OR TableB._hash IS NULL
            """.format(
                dataset_name=bq_dataset.dataset_id,
                **self.__get_table_names())

            src_query_parameters = (
                bigquery.ScalarQueryParameter('commit_sha', 'STRING', commit_sha),
                bigquery.ScalarQueryParameter('ts', 'TIMESTAMP', ts),
            )

            metadata_table = self._get_table_ref(self.METADATA_TABLE_NAME)
            job = self._async_copy_table_data(src_query, src_query_parameters, metadata_table)

            return BqJob(job)

    def _query_head_data(self, sha_list):
        from google.cloud import bigquery
        from google.cloud.bigquery import QueryJobConfig

        bq_client = self._connection

        metadata_fields, staging_metadata_fields = self.__get_metadata_fields_for_union()

        with bq_client.get_cursor() as bq_dataset:
            query = """
            #standardSQL
            #_query_head_data
            SELECT * EXCEPT(_max_sha, _max_ts, _ts)
            FROM (
              SELECT {staging_metadata_fields}
              FROM `{dataset_name}.{metadata_staging_index_table_name}`
              WHERE _sha IN UNNEST(@sha_list)
              UNION ALL
              SELECT {metadata_fields}
              FROM `{dataset_name}.{metadata_table_name}`
              WHERE _sha IN UNNEST(@sha_list)) AS metadata_staging_combine
              INNER JOIN (
                  SELECT  _sha AS _max_sha, MAX(_ts) AS _max_ts
                  FROM `{dataset_name}.{metadata_staging_index_table_name}`
                  WHERE _sha IN UNNEST(@sha_list)
                  GROUP BY _sha
                  UNION ALL
                  SELECT  _sha AS _max_sha, MAX(_ts) AS _max_ts
                  FROM `{dataset_name}.{metadata_table_name}`
                  WHERE _sha IN UNNEST(@sha_list)
                  GROUP BY _sha
              ) _max_metadata
              ON metadata_staging_combine._sha = _max_sha
              WHERE metadata_staging_combine._ts = _max_ts;
            """.format(
                dataset_name=bq_dataset.dataset_id,
                staging_metadata_fields=staging_metadata_fields,
                metadata_fields=metadata_fields,
                **self.__get_table_names())

            query_parameters = (
                bigquery.ArrayQueryParameter('sha_list', 'STRING', sha_list),
            )

            job_config = QueryJobConfig()
            job_config.query_parameters = query_parameters

            items_iter, _ = self._query_sync(query, job_config, process_row=self.build_dict)

            return items_iter

    @classmethod
    def __table_fields(cls, table_ref):
        for field in table_ref:
            yield field.name

    @classmethod
    def __union_fields(cls, fields_a, fields_b):
        def field_value(field_name, missing_fields=None):
            if field_name in (missing_fields or []):
                return 'NULL as `%s`' % field_name

            return '`%s`' % field_name

        new_fields = list(set(fields_a) - set(fields_b))

        fields_a_select = ','.join([field_value(field) for field in fields_a])
        fields_b_select = ','.join([field_value(field, new_fields) for field in fields_a])

        return fields_b_select, fields_a_select

    def __get_metadata_fields_for_union(self, except_fields=None):
        self.__using_staging_table()
        self.__using_metadata_table()

        staging_metadata_fields = set(list(self.__table_fields(self.__staging_table.schema))) - set(except_fields or [])
        metadata_fields = set(list(self.__table_fields(self.__metadata_table.schema))) - set(except_fields or [])

        return self.__union_fields(staging_metadata_fields, metadata_fields)

    def _query(self, sql_vars, select_fields, where, result_url=None):
        import google.cloud.exceptions
        from google.cloud import bigquery
        from google.cloud.bigquery import QueryJobConfig

        bq_client = self._connection
        sql_vars['random_function'] = '_phr'

        user_limit = sql_vars.get('limit')
        user_limit = 'LIMIT %s' % user_limit if user_limit is not None else ''

        system_limit = '' if user_limit else 'LIMIT 1000' if result_url is None else ''

        offset = sql_vars.get('offset')
        offset = 'OFFSET %s' % offset if offset is not None else ''

        if sql_vars.get('split_field') is None:
            query_phase = """
                    CASE
                      WHEN _phr >= $sample_percentile + $phase_train_start * $sample AND _phr < $sample_percentile + $phase_train_end * $sample THEN 'train'
                      WHEN _phr >= $sample_percentile + $phase_test_start * $sample AND _phr < $sample_percentile + $phase_test_end * $sample THEN 'test'
                      WHEN _phr >= $sample_percentile + $phase_validation_start * $sample AND _phr < $sample_percentile + $phase_validation_end * $sample THEN 'validation'
                      ELSE NULL
                    END as _phase            
            """
        else:
            query_phase = '%s as _phase' % sql_vars.get('split_field')

        def query_size_staging_without_where():
            return """
                #standardSQL
                #query_size_staging_without_where
                SELECT SUM(_size) AS `size`, COUNT(_size) AS `count`
                FROM (
                    SELECT *
                    FROM (
                        SELECT size as _size, name as _sha, ROW_NUMBER() OVER (PARTITION BY name ORDER BY ts DESC) row_number                                  
                        FROM {dataset_name}.{index_staging_index_table_name}
                    )
                    WHERE row_number = 1
                    {user_limit}
                )
             """

        def query_size_without_staging_without_where():
            return """
                #standardSQL
                #query_size_without_staging_without_where
                SELECT SUM(_size) AS `size`, COUNT(_size) AS `count`
                FROM (
                    SELECT *
                    FROM (
                        SELECT size as _size, name as _sha, ROW_NUMBER() OVER (PARTITION BY name ORDER BY ts DESC) row_number                                  
                        FROM {dataset_name}.{index_table_name}
                        WHERE ts <= @version_ts
                    )
                    WHERE row_number = 1
                    {user_limit}
                )
             """

        def query_size_staging():
            if where is None:
                return query_size_staging_without_where()

            metadata_fields, staging_metadata_fields = self.__get_metadata_fields_for_union({'_hash', 'url'})

            return """
                #standardSQL
                #query_size_staging
                SELECT SUM(_size) as `size`, COUNT(_size) as `count`
                FROM (
                    SELECT  _sha, _size
                    FROM ( # Bring all the items from staging index 
                      SELECT *
                      FROM (
                        SELECT name as _sha, _size 
                        FROM (
                          SELECT name, size as _size, ROW_NUMBER() OVER (PARTITION BY name ORDER BY ts DESC) row_number
                          FROM {dataset_name}.{index_staging_index_table_name}
                         )
                        WHERE row_number = 1
                      ) staging_index
                      LEFT JOIN (
                        SELECT _sha
                        FROM (
                          SELECT _sha, ROW_NUMBER() OVER (PARTITION BY _sha ORDER BY _ts DESC) row_number
                          FROM {dataset_name}.{metadata_staging_index_table_name}
                        )
                        WHERE row_number = 1
                      ) staging_meta
                      USING(_sha) # _sha is actual a name here
                    ) AS staging_items
                    LEFT JOIN (
                      # join this with latest (_ts) metadata
                      SELECT * EXCEPT(_commit_sha, row_number)
                      FROM (
                        SELECT *, ((FARM_FINGERPRINT({phase_seed}) + POW(2, 63)) / POW(2, 64)) as _phr
                          FROM (
                            SELECT *, ROW_NUMBER() OVER (PARTITION BY _sha ORDER BY _ts DESC) row_number
                            FROM  (
                              SELECT {staging_metadata_fields}
                              FROM {dataset_name}.{metadata_staging_index_table_name}
                              UNION ALL
                              SELECT {metadata_fields}
                              FROM {dataset_name}.{metadata_table_name}
                            )
                          )
                          WHERE row_number = 1
                      )
                    ) AS meta
                    USING(_sha)
                    WHERE({where})
                    {user_limit}                    
                )
             """.replace('{staging_metadata_fields}', staging_metadata_fields).replace('{metadata_fields}', metadata_fields)

        def query_staging():
            metadata_fields, staging_metadata_fields = self.__get_metadata_fields_for_union({'_hash', 'url'})

            return """
                #standardSQL
                # query_staging
                SELECT * EXCEPT(_ts, _phr), {query_phase}
                FROM (
                    SELECT  *
                    FROM ( # Bring all the items from staging index 
                      SELECT *
                      FROM (
                        SELECT name as _sha, _size 
                        FROM (
                          SELECT name, size as _size, ROW_NUMBER() OVER (PARTITION BY name ORDER BY ts DESC) row_number
                          FROM {dataset_name}.{index_staging_index_table_name}
                         )
                        WHERE row_number = 1
                      ) staging_index
                      LEFT JOIN (
                        SELECT _sha
                        FROM (
                          SELECT _sha, ROW_NUMBER() OVER (PARTITION BY _sha ORDER BY _ts DESC) row_number
                          FROM {dataset_name}.{metadata_staging_index_table_name}
                        )
                        WHERE row_number = 1
                      ) staging_meta
                      USING(_sha) # _sha is actual a name here
                    ) AS staging_items
                    LEFT JOIN (
                      # join this with latest (_ts) metadata
                      SELECT * EXCEPT(_commit_sha, row_number)
                      FROM (
                        SELECT *
                          FROM (
                            SELECT *, ROW_NUMBER() OVER (PARTITION BY _sha ORDER BY _ts DESC) row_number
                            FROM  (
                              SELECT {staging_metadata_fields}
                              FROM {dataset_name}.{metadata_staging_index_table_name}
                              UNION ALL
                              SELECT {metadata_fields}
                              FROM {dataset_name}.{metadata_table_name}
                            )
                          )
                          WHERE row_number = 1
                      )
                      RIGHT JOIN (
                        # This join will give us the latest URL (_url) and hash (_hash) of the data (by name)
                        SELECT ((FARM_FINGERPRINT({phase_seed}) + POW(2, 63)) / POW(2, 64)) as _phr, *
                        FROM (
                          SELECT name as _sha, sha as _hash, _url
                          FROM ( # This will remove any duplicates we might have
                            SELECT * EXCEPT(row_number)
                            FROM (
                              SELECT *, ROW_NUMBER() OVER (PARTITION BY name ORDER BY ts DESC) row_number
                              FROM (
                                SELECT name, ts, url as _url, sha
                                FROM {dataset_name}.{index_staging_index_table_name}
                                UNION ALL
                                SELECT name, ts, url as _url, sha
                                FROM {dataset_name}.{index_table_name}
                                UNION ALL
                                SELECT _sha as name, NULL as ts, NULL as  _url, NULL as sha
                                FROM {dataset_name}.{metadata_staging_index_table_name}            
                              )
                            )
                            WHERE row_number = 1                                                        
                          )
                        )
                      ) index_with_hash
                      USING(_sha)
                    ) AS meta
                    USING(_sha)
                )
                WHERE({where})
                ORDER BY _ts DESC, _sha
                {user_limit_or_system_limit}
                {offset}
             """.replace('{staging_metadata_fields}', staging_metadata_fields).replace('{metadata_fields}', metadata_fields)

        def query_size_without_staging():
            if where is None:
                return query_size_without_staging_without_where()

            return """
                #standardSQL
                #query_size_without_staging
                SELECT SUM(_size) AS `size`, COUNT(_size) AS `count`
                FROM (
                    SELECT *
                    FROM (
                        SELECT *, {query_phase}
                        FROM (
                            SELECT *, ((FARM_FINGERPRINT({phase_seed}) + POW(2, 63)) / POW(2, 64)) as _phr                    
                            FROM (
                                SELECT *, ROW_NUMBER() OVER (PARTITION BY _sha ORDER BY _ts DESC) row_number
                                FROM (
                                  SELECT * EXCEPT(row_number, _hash)
                                  FROM (
                                      SELECT *, ROW_NUMBER() OVER (PARTITION BY _sha ORDER BY _ts DESC) as row_number
                                      FROM {dataset_name}.{metadata_table_name}
                                      WHERE _ts <= @version_ts
                                    )
                                    WHERE row_number = 1
                                )
                                RIGHT JOIN (
                                    SELECT * EXCEPT(row_number)
                                    FROM (
                                        SELECT size as _size, name as _sha, ROW_NUMBER() OVER (PARTITION BY name ORDER BY ts DESC) row_number                                  
                                        FROM {dataset_name}.{index_table_name}
                                        WHERE ts <= @version_ts
                                    )
                                    WHERE row_number = 1
                                )
                                USING(_sha)
                            )
                        )
                    )
                    WHERE row_number = 1 AND ({where})
                    {user_limit}
                )
             """

        def query_all_meta_data_without_staging():
            return """
                #standardSQL
                #query_all_meta_data_without_staging
                SELECT * EXCEPT(_ts, _phr, _commit_sha, commit_sha), commit_sha AS _commit_sha
                FROM (
                    SELECT {query_phase}, *                    
                    FROM (
                        SELECT *, ((FARM_FINGERPRINT({phase_seed}) + POW(2, 63)) / POW(2, 64)) as _phr
                        FROM (
                            # This join will give us the latest URL (url) and hash (_hash) of the data (by name)
                            # This will remove any duplicates we might have
                            SELECT * EXCEPT(row_number)
                            FROM (
                                SELECT size as _size, commit_sha, name as _sha, sha as _hash, url as _url, 
                                       ROW_NUMBER() OVER (PARTITION BY name ORDER BY ts DESC) row_number                                  
                                FROM {dataset_name}.{index_table_name}
                                WHERE ts <= @version_ts
                            )
                            WHERE row_number = 1
                        )
                        LEFT JOIN (
                          SELECT * EXCEPT(row_number, _hash)
                          FROM (
                            SELECT *, ROW_NUMBER() OVER (PARTITION BY _sha ORDER BY _ts DESC) as row_number
                            FROM {dataset_name}.{metadata_table_name}
                            WHERE _ts <= @version_ts
                          )
                          WHERE row_number = 1                        
                        )
                        USING(_sha)                        
                    )
                    WHERE _phr is not NULL
                )
                WHERE ({where})
                {user_limit_or_system_limit}
                {offset}
             """

        def get_phase_seed():
            fields = []
            for key in [group_key, datapoint_key]:
                if not key:
                    continue

                fields.append('CAST(`{group}` AS STRING)'.format(group=key))

            if not fields:
                fields.append('`_sha`')

            fields.append('"$seed"')

            return "CONCAT({fields})".format(fields=','.join(fields))

        def append_ver_if_needed(name, var_type, src_query_parameters):
            if '@' + name in query_sql:
                version_ts = sql_vars[name]
                src_query_parameters.append(bigquery.ScalarQueryParameter(name, var_type, version_ts))

            return src_query_parameters

        def handle_bad_query_request(bad_request):
            field = self._field_not_found(bad_request.message)

            if field is not None:
                raise MetadataFieldNotFound(field)

        def format_query(query):
            return query.format(
                dataset_name=bq_dataset.dataset_id,
                phase_seed=phase_seed,
                where=where or 'True',
                select=','.join(select_fields),
                user_limit_or_system_limit=user_limit or system_limit,
                system_limit=system_limit,
                user_limit=user_limit,
                offset=offset,
                query_phase=query_phase,
                **self.__get_table_names())

        def return_iter(query_query_job):
            data_iter, _ = self._query_sync_by_job(query_query_job, process_row=self.build_dict)

            return data_iter

        def get_size_result_job(job_config):
            def process_size_result(results):
                for result in results:
                    return result['size'], result['count']

            return BqJob(
                self._query_async(size_query, job_config), process_size_result)

        def create_job_config(current_result_url=None):
            src_query_parameters = append_ver_if_needed('version_ts', 'TIMESTAMP', [])

            job_config = QueryJobConfig()
            job_config.query_parameters = src_query_parameters
            job_config.use_query_cache = True

            if current_result_url is not None:
                job_config.allow_large_results = True
                job_config.destination = self._get_temp_table(prefix='query_data')

            return job_config

        with bq_client.get_cursor() as bq_dataset:
            version_var = sql_vars.get('version')

            group_key = sql_vars.get('group')
            datapoint_key = sql_vars.get('datapoint')

            phase_seed = get_phase_seed()

            if version_var is not None and version_var.lower() == 'staging':
                query_sql, size_query = query_staging(), query_size_staging()
            else:
                query_sql, size_query = query_all_meta_data_without_staging(), query_size_without_staging()

            query_sql = self.fill_in_vars(format_query(query_sql), sql_vars)
            size_query = self.fill_in_vars(format_query(size_query), sql_vars)

            def do_query():
                query_job_config = create_job_config(result_url)
                full_result_query_job = self._query_async(query_sql, query_job_config)

                query_job = get_size_result_job(job_config=create_job_config())

                if result_url is None:
                    total_size, total_count = query_job.wait()

                    return return_iter(full_result_query_job), total_count, total_size

                full_result_query_job.result()

                extract_job = self._extract_table(query_job_config.destination, result_url)

                BqJob(extract_job).wait()
                total_size, total_count = query_job.wait()

                return None, total_count, total_size

            # noinspection PyUnresolvedReferences
            try:
                return do_query()
            except google.cloud.exceptions.BadRequest as ex:
                handle_bad_query_request(ex)
                logging.exception('bad query')
                raise MetadataOperationError()

    @classmethod
    def _field_not_found(cls, message):
        field_not_found_re = r'Unrecognized name:\ (?P<field>.*) at \[\d+:\d+\]'

        m = re.match(field_not_found_re, message)

        return None if m is None else m.group("field")

    def get_all_data(self, sha):
        pass

    def get_data_for_commit(self, sha, commit_sha):
        pass

    def delete_all(self):
        metadata_table = self._get_table_name(self._connection.table_prefix, self.METADATA_TABLE_NAME)
        staging_metadata_table_prefix = self._get_table_name(self._connection.table_prefix, self.STAGING_TABLE_NAME)
        self._connection.delete_tables([metadata_table, staging_metadata_table_prefix])

    def get_commit_statistics(self, commit_sha, most_frequent_values_limit=100):
        from google.cloud.bigquery import QueryJobConfig
        import google.cloud.exceptions

        def get_metadata_fields():
            metadata_table_ref = self._get_table_ref(self.METADATA_TABLE_NAME)
            # noinspection PyUnresolvedReferences
            try:
                metadata_table = bq_client.get_table(metadata_table_ref)
            except google.cloud.exceptions.NotFound:
                logging.info('table %s not found', self.METADATA_TABLE_NAME)
                return None

            default_schema = self._default_table_schema()
            return [field.name for field in metadata_table.schema if field not in default_schema]

        def query_top_values(current_metadata_field):
            query = """
                #standardSQL
                #query_top_values
                SELECT `{field}`, COUNT(*) as frequency
                FROM (
                  SELECT {field}, ROW_NUMBER() OVER (PARTITION by _sha ORDER BY _ts DESC) row_number
                  FROM `{dataset_name}.{metadata_table_name}`
                  WHERE  _ts <= (
                    SELECT MAX(_ts)
                    FROM `{dataset_name}.{metadata_table_name}`
                    WHERE `_commit_sha` = '{commit_sha}')
                )
                WHERE row_number = 1
                GROUP BY `{field}`
                ORDER BY frequency DESC
                LIMIT {limit}
            """.format(
                dataset_name=bq_dataset.dataset_id,
                field=current_metadata_field,
                commit_sha=commit_sha,
                limit=most_frequent_values_limit,
                **self.__get_table_names())
            return self._query_async(query, QueryJobConfig())

        bq_client = self._connection
        with bq_client.get_cursor() as bq_dataset:
            metadata_fields = get_metadata_fields()

            if metadata_fields is None:
                return {}

            query_jobs = {field: query_top_values(field) for field in metadata_fields}

            metadata_statistics = {}
            for metadata_field, query_job in query_jobs.items():
                top_values = [tuple(row) for row in query_job.result()]
                metadata_statistics[metadata_field] = top_values

        return metadata_statistics

    def __get_table_names(self):
        bq_client = self._connection

        metadata_table_name = self._get_table_name(bq_client.table_prefix, self.METADATA_TABLE_NAME)
        metadata_staging_index_table_name = self._get_table_name(bq_client.table_prefix, self.__get_staging_table_name())
        index_staging_index_table_name = self._get_table_name(bq_client.table_prefix, self.__get_staging_index_table_name())
        index_table_name = self._get_table_name(bq_client.table_prefix, self.INDEX_TABLE_NAME)

        return {
            'metadata_table_name': metadata_table_name,
            'metadata_staging_index_table_name': metadata_staging_index_table_name,
            'index_staging_index_table_name': index_staging_index_table_name,
            'index_table_name': index_table_name,
        }

    def rpc_total_count_items(self):
        bq_client = self._connection

        with bq_client.get_cursor() as bq_dataset:
            query = """
                #standardSQL
                #rpc_total_count_items
                SELECT *
                FROM (
                    SELECT 'item' as _id, COUNT(DISTINCT name) as total_data_points
                    FROM (
                      SELECT DISTINCT name FROM `{dataset_name}.{index_table_name}` 
                      UNION ALL
                      SELECT DISTINCT name FROM `{dataset_name}.{index_staging_index_table_name}`
                    )
                )
                LEFT JOIN (
                  SELECT 'item' as _id, COUNT(DISTINCT name) as version_total_data_points
                  FROM `{dataset_name}.{index_staging_index_table_name}`                
                )
                USING (_id)
                LEFT JOIN (
                    SELECT 'item' as _id, COUNT(DISTINCT _sha) as total_meta_data
                    FROM (
                      SELECT DISTINCT _sha FROM `{dataset_name}.{metadata_table_name}`
                      UNION ALL
                      SELECT DISTINCT _sha FROM `{dataset_name}.{metadata_staging_index_table_name}`
                    )                
                )
                USING (_id)
                LEFT JOIN (
                  SELECT 'item' as _id, COUNT(DISTINCT _sha) as version_total_meta_data
                  FROM `{dataset_name}.{metadata_staging_index_table_name}`                
                )
                USING (_id)                                               
                """.format(dataset_name=bq_dataset.dataset_id, **self.__get_table_names())

            return self._return_async_dict(
                query,
                'total_data_points', 'version_total_data_points',
                'total_meta_data', 'version_total_meta_data')

    def rpc_get_commit_id(self):
        bq_client = self._connection

        with bq_client.get_cursor() as bq_dataset:
            src_query = """ 
                #standardSQL
                #rpc_get_commit_id
                SELECT *
                FROM (
                    SELECT 'item' as _id, TO_HEX(SHA1(STRING_AGG(_hash ORDER BY _sha))) AS meta_sha
                    FROM (
                        SELECT * EXCEPT(row_number)
                        FROM (
                            SELECT _hash, _sha, ROW_NUMBER() OVER (PARTITION BY _sha ORDER BY _ts DESC) row_number
                            FROM (
                              SELECT _hash, _ts, _sha FROM `{dataset_name}.{metadata_staging_index_table_name}`
                              UNION ALL
                              SELECT _hash, _ts, _sha FROM `{dataset_name}.{metadata_table_name}`
                          )
                        )
                        WHERE row_number=1
                    )
                )
                LEFT JOIN ( 
                    SELECT 'item' as _id, TO_HEX(SHA1(STRING_AGG(sha ORDER BY name))) AS index_sha
                    FROM (
                      SELECT * EXCEPT(row_number)
                      FROM (
                        SELECT sha, name, ROW_NUMBER() OVER (PARTITION BY name ORDER BY ts DESC) row_number
                        FROM (
                          SELECT sha, ts, name FROM `{dataset_name}.{index_staging_index_table_name}`
                          UNION ALL
                          SELECT sha, ts, name FROM `{dataset_name}.{index_table_name}`
                        )
                      )
                      WHERE row_number=1
                    )       
                )
                USING (_id)
              """.format(dataset_name=bq_dataset.dataset_id, **self.__get_table_names())

        return self._return_async_dict(src_query, 'index_sha', 'meta_sha')

    def delete_version(self, version_id):
        from google.cloud import bigquery

        query_parameters = (
            bigquery.ScalarQueryParameter('version_id', 'STRING', version_id),
        )

        metadata_table_ref = self._get_table_ref(self.METADATA_TABLE_NAME)

        with self._connection.get_cursor() as bq_dataset:
            query = '''
                #standardSQL
                #delete_version
                SELECT * 
                FROM `{dataset_name}.{metadata_table_name}`
                WHERE `_commit_sha` <> @version_id 
            '''.format(
                dataset_name=bq_dataset.dataset_id,
                metadata_table_name=metadata_table_ref.table_id,
            )

            return self.override_table(query, query_parameters, metadata_table_ref)



# -*- coding: utf8 -*-
import os
import shutil
import errno
import re
import datetime
import six
import hashlib
from .exceptions import AccessDenied

try:
    from xattr import xattr as ExtendedFileAttributes
    has_xattr_support = True
except ImportError:
    has_xattr_support = False
    ExtendedFileAttributes = None


ignore_files = ['.DS_Store']


def __file_info_hash(sys, etag_or_filename):
    sha_1 = hashlib.sha1()

    if sys == 'local':
        with open(etag_or_filename, 'rb') as f:
            sha_1.update(f.read())
    else:
        etag = etag_or_filename.encode('ascii')
        sha_1.update(etag)

    return sha_1.hexdigest()


def __s3_info_to_file_info(item, fields=None):
    params = {
        'path': item['Key'],
        'sys': 's3',
    }

    if fields is None or 'size' in fields:
        params['size'] = item['Size']

    etag = item['ETag'].replace('"', '')

    if fields is None or 'etag' in fields:
        params['etag'] = etag

    if fields is None or 'mtime' in fields:
        params['mtime'] = item['LastModified'].replace(tzinfo=None)

    if fields is None or 'sha' in fields:
        params['sha'] = __file_info_hash(params['sys'], etag)

    params['url'] = '{moniker}{bucket}/{Key}'.format(**item)

    return params


def __create_s3_paginator(**kwargs):
    import boto3

    client = boto3.client('s3')
    paginator = client.get_paginator('list_objects_v2')

    return paginator.paginate(**kwargs)


def __split_bucket_prefix(bucket):
    try:
        bucket, prefix = bucket.split('/', 1)
    except ValueError:
        bucket, prefix = bucket, None

    return bucket, prefix


def __enumerate_path_s3(path, fields=None):
    from botocore.exceptions import ClientError

    path = remove_moniker(path)

    bucket, prefix = __split_bucket_prefix(path)

    try:
        for page in __create_s3_paginator(Bucket=bucket, Prefix=prefix):
            for item in page.get('Contents', []):
                if item['Key'].endswith('/'):  # skip folders
                    continue

                item['bucket'] = bucket
                item['moniker'] = s3_moniker

                yield __s3_info_to_file_info(item, fields)
    except ClientError as ex:
        if ex.response.get('Error', {}).get('Code') == 'AccessDenied':
            raise AccessDenied('Access denied accessing bucket "{}"'.format(bucket))

        raise


def bucket_print_name(path):
    if path.startswith(s3_moniker):
        import boto3
        from botocore.exceptions import ClientError

        path = remove_moniker(path)
        bucket_name = path.split('/')[0]

        try:
            client = boto3.client('s3')

            response = client.get_bucket_location(Bucket=bucket_name)

            location = response.get('LocationConstraint') or 'us-east-1'
            return '%s (%s)' % (path, location)
        except ClientError:
            pass

    return path


def __get_xattr(attr):
    if not attr:
        return None, None

    try:
        data = attr.get('ai.ml.sha')
        data = data.decode()
        data = data.split('|')
        data[0] = float(data[0])
    except IOError as ex:
        data = (None, None)

    return data


def __set_xattr(attr, mtime, sha):
    global has_xattr_support

    if not attr:
        return

    data_encoded = '%s|%s' % (mtime, sha)
    data_encoded = data_encoded.encode()
    try:
        attr.set('ai.ml.sha', data_encoded)
    except IOError as ex:
        if ex.errno == 95:
            has_xattr_support = False
            return

        raise


def __file_to_info(path, st, fields):
    attr = ExtendedFileAttributes(path) if has_xattr_support else None

    params = {
        'path': path,
        'sys': 'local',
        'mtime': datetime.datetime.fromtimestamp(st.st_mtime),
        'ctime': datetime.datetime.fromtimestamp(st.st_ctime),
        'size': st.st_size,
        'mode': st.st_mode
    }

    if fields is None or 'sha' in fields:
        attr_mtime, attr_sha = __get_xattr(attr)

        if attr_mtime != st.st_mtime:
            params['sha'] = __file_info_hash(params['sys'], path)
            __set_xattr(attr, st.st_mtime, params['sha'])
        else:
            params['sha'] = attr_sha

    return params


def __file_name_to_info(filename, fields=None):
    return __file_to_info(filename, os.stat(filename), fields)


def __file_entry_to_info(entry, fields=None):
    return __file_to_info(entry.path, entry.stat(), fields)


def __recursive_scan(current_path, fields):
    try:
        from os import scandir
    except ImportError:
        from scandir import scandir

    try:
        for entry in scandir(current_path):
            if entry.is_dir():
                for info in __recursive_scan(entry.path, fields):
                    yield info

                continue

            yield __file_entry_to_info(entry, fields)
    except OSError as ex:
        if ex.errno == 20:
            yield __file_name_to_info(current_path, fields)
            return

        raise


def __enumerate_path_local(path, fields=None):
    for info in __recursive_scan(path, fields):
        yield info


s3_moniker = 's3://'


def enumerate_path_with_info(path, fields=None):
    if path.startswith(s3_moniker):
        enumerate_method = __enumerate_path_s3
    else:
        enumerate_method = __enumerate_path_local

    for file_info in enumerate_method(path, fields):
        full_object_name = file_info['path']
        file_name = os.path.basename(full_object_name)

        if file_name in ignore_files:
            continue

        yield file_info


def enumerate_paths_with_info(paths, fields=None):
    if isinstance(paths, six.string_types):
        paths = [paths]

    for path in paths:
        for file_info in enumerate_path_with_info(path, fields=fields):
            yield file_info


def enumerate_paths(paths):
    for file_info in enumerate_paths_with_info(paths, fields=['size']):
        yield file_info['path']


def get_total_files_in_path(paths, callback=None):
    total_files = 0

    for file_name in enumerate_paths(paths):
        count_file = True
        if callback is not None:
            count_file = callback(file_name)

        if count_file:
            total_files += 1

    return total_files


def is_glob(path):
    return '*' in path or '?' in path


def has_var(path):
    return '$' in path


def __validate_path_if_needed(path, validate_path):
    if not is_glob(path) and validate_path and not os.path.exists(path):
        raise IOError()


def __unslash(path):
    if path.endswith(os.sep):
        path = os.path.join(path, '')

    return path


def expend_and_validate_path(path, expand_vars=True, validate_path=True, abs_path=True):
    if path is None:
        return path

    result_path = path

    if expand_vars:
        result_path = os.path.expandvars(result_path)

    if has_moniker(path):
        return result_path

    result_path = os.path.expanduser(result_path)

    if abs_path:
        result_path = os.path.abspath(result_path)

    __validate_path_if_needed(result_path, validate_path)

    result_path = __unslash(result_path)

    return result_path


def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i + n]


def get_batch_of_files_from_paths(file_names, batch_size):
    for batch in chunks(file_names, batch_size):
        yield batch


def safe_make_dirs(dir_name):
    try:
        os.makedirs(dir_name)
    except OSError as ex:
        if ex.errno != errno.EEXIST:
            raise


def path_elements(path):
    if path.endswith(os.sep):
        path = path[:-1]

    folders = []
    while True:
        path, folder = os.path.split(path)

        if len(folder) > 0:
            folders.append(folder)
            continue

        if len(path) > 0:
            folders.append(path)

        break

    folders.reverse()

    return folders


def safe_rm_tree(path):
    try:
        shutil.rmtree(path)
    except OSError as ex:
        if ex.errno != errno.ENOENT:
            raise


class DestPathEnum(object):
    @classmethod
    def find_root(cls, dest):
        elements = path_elements(dest)

        root = []
        for element in elements:
            if has_var(element):
                break

            root.append(element)

        return os.path.join(*root)

    @classmethod
    def get_path_vars(cls, pattern, path):
        if path is None:
            return {}

        path_no_ext, file_extension = os.path.splitext(path)

        # in case the user has already specify '.' in the ext don't use dot in the var
        if '.$ext' in pattern or '.$@ext' in pattern:
            file_extension = file_extension[1:]

        return {
            'name': os.path.basename(path),
            'dir': os.path.dirname(path),
            'base_name': os.path.basename(path_no_ext),
            'ext': file_extension,
            'extension': file_extension
        }

    @classmethod
    def ___add_sys_var(cls, name, value, current_vars):
        current_vars['@' + name] = value

        if name not in current_vars:
            current_vars[name] = value

    @classmethod
    def __fill_in_vars(cls, path, replace_vars):
        replace_vars_keys = sorted(replace_vars.keys(), reverse=True)

        for var_name in replace_vars_keys:
            var_value = replace_vars[var_name]
            path = path.replace('$' + var_name, str(var_value))
            path = path.replace('#' + var_name, str(var_value))

        return path

    @classmethod
    def get_full_path(cls, dest, item):
        for key, val in cls.get_path_vars(dest, item.get('@path')).items():
            cls.___add_sys_var(key, val, item)

        phase = item.get('@phase')
        cls.___add_sys_var('phase', phase, item)
        item['@'] = phase

        dest_file = cls.__fill_in_vars(dest, item)

        return dest_file

    @classmethod
    def get_dest_path(cls, dest_folder, dest_file):
        if not dest_file:
            dest_file = '$@name'

        return os.path.join(dest_folder, dest_file)


def create_dir(dirname):
    if not dirname or os.path.exists(dirname):
        return

    try:
        os.makedirs(dirname)
    except OSError as exc:
        if exc.errno != errno.EEXIST:
            raise


def remove_dir(dirname):
    try:
        shutil.rmtree(dirname)
    except (OSError, IOError):
        pass


def purge(dir, pattern):
    for f in os.listdir(dir):
        if re.search(pattern, f):
            os.remove(os.path.join(dir, f))


def flatten_dir(root_dir):
    root_walk = os.walk(root_dir)
    _, top_level_dirs, _ = next(root_walk)

    for dirpath, _, filenames in root_walk:
        for filename in filenames:
            filepath = os.path.join(dirpath, filename)
            shutil.move(filepath, root_dir)

    for top_level_dir in top_level_dirs:
        remove_dir(os.path.join(root_dir, top_level_dir))


def has_moniker(name):
    return '://' in name


def remove_moniker(name):
    try:
        index = name.index('://')
        return name[index + 3:]
    except ValueError:
        return name



# -*- coding: utf8 -*-
import json
import click
import click.types
from .legit.context import init_context2


class DataVolumeIdParamType(click.types.IntParamType):
    # noinspection PyMethodMayBeStatic
    def complete(self, ctx, _incomplete):
        if ctx.obj is None:
            init_context2(ctx)

        projects = ctx.obj.handle_api(ctx.obj, ctx.obj.session, 'get', 'data_volumes')

        result = []
        for project in projects.get('volumes', []):
            project_id = str(project['id'])

            display_name = project.get('display_name')

            org = project.get('org') or 'me'

            display_name += ' @' + org

            result.append((project_id, display_name))

        return result


class ProjectIdParamType(click.types.IntParamType):
    # noinspection PyMethodMayBeStatic
    def complete(self, ctx, _incomplete):
        if ctx.obj is None:
            init_context2(ctx)

        projects = ctx.obj.handle_api(ctx.obj, ctx.obj.session, 'get', 'projects')

        result = []
        for project in projects.get('projects', []):
            project_id = str(project['project_id'])

            display_name = project.get('display_name')

            org = project.get('org') or 'me'

            display_name += ' @' + org

            result.append((project_id, display_name))

        return result


def processes_option():
    def decorator(f):
        return click.option('--processes', default=-1, type=int, required=False)(f)

    return decorator


def no_progressbar_option():
    def decorator(f):
        return click.option('--no_progressbar/--enable_progressbar', default=False, is_flag=True, required=False)(f)

    return decorator


def data_volume_id_argument():
    def decorator(f):
        return click.argument(
            'volumeId', type=DataVolumeIdParamType(), envvar='VOLUMEID')(f)

    return decorator


def alert_type_option(required=False, multiple=False):
    all_options = ['started', 'stopped', 'ended', 'failed', 'dataWarning']

    def validate_alert_type(ctx, option, value):
        value = list(value)

        for alert_type in value:
            if alert_type == 'all':
                value.extend(all_options)

        value = [v for v in value if v in all_options]

        return list(set(value))

    def decorator(f):
        return click.option(
            '--alertType', '-a', callback=validate_alert_type, type=click.Choice(['all'] + all_options), required=required, multiple=multiple,
            help='Alert type for project subscription')(f)

    return decorator


def project_id_option(required=False, multiple=False):
    def decorator(f):
        return click.option(
            '--projectId', '-p', type=ProjectIdParamType(), required=required, multiple=multiple,
            help='The project Id. Use `mali projects list` to find your project Ids')(f)

    return decorator


def experiment_id_option(required=False):
    def decorator(f):
        return click.option(
            '--experimentId', '-e', type=int, metavar='<int>', required=required,
            help='The experiment ID. Use `mali experiments list` to find your experiment IDs')(f)

    return decorator


def chart_scope_option(required=False):
    def decorator(f):
        return click.option(
            '--chartScope', '-cs', type=click.Choice(['test', 'validation', 'train']), required=required,
            default='test', help='Scope type.')(f)

    return decorator


def chart_type_option(required=False):
    def decorator(f):
        return click.option(
            '--chartType', '-ct', type=click.Choice(['line']), required=required,
            default='line', help='Graph type.')(f)

    return decorator


def chart_name_option(required=False):
    def decorator(f):
        return click.option(
            '--chartName', '-c', metavar='<str>', required=required,
            help='The name of the chart. The name is used in order to identify the chart against different '
                 'experiments and through the same experiment.')(f)

    return decorator


def chart_x_option(required=False):
    def decorator(f):
        return click.option(
            '--chartX', '-cx', metavar='<json_string>', required=required,
            help='Array of m data points (X axis), Can be Strings, Integers or Floats.')(f)

    return decorator


def chart_y_option(required=False):
    def decorator(f):
        return click.option(
            '--chartY', '-cy', metavar='<json_string>', required=required,
            help='Array/Matrix of m data values. Can be either array m of Integers/Floats or a matrix (m arrays having n Ints/Floats each),  a full matrix describing the values of every chart in every data point')(f)

    return decorator


def chart_y_name_option(required=False):
    def decorator(f):
        return click.option(
            '--chartYName', '-cyn', metavar='<json_str>', required=required, default='Y',
            help='Display name for chart(s) Y axis')(f)

    return decorator


def chart_x_name_option(required=False):
    def decorator(f):
        return click.option(
            '--chartXName', '-cxn', metavar='<str>', required=required, default='X',
            help='Display name for charts X axis')(f)

    return decorator


def metrics_option(required=False):
    def decorator(f):
        return click.option(
            '--metrics', '-m', metavar='<json_string>', required=required,
            help='The metrics of the experiment as a jsonified string. The key should be the metric '
                 'name with "ex" prefix e.g. "ex_cost". The value is the metric value in String, Float, '
                 'Integer or Boolean.')(f)

    return decorator


def model_weights_hash_option(required=False):
    def decorator(f):
        return click.option(
            '--weightsHash', '-wh', metavar='<sha1_hex>', required=required,
            help="The hexadecimal sha1 hash of the model's weights")(f)

    return decorator


def validate_json(ctx, param, value):
    try:
        if value is None:
            return None

        return json.loads(value)
    except ValueError:
        raise click.BadParameter('not valid json', param=param, ctx=ctx)



# -*- coding: utf8 -*-
import collections
import json
import logging
import sys

import time
from requests import HTTPError
from six.moves.urllib import parse

from .eprint import eprint
from .config import get_prefix_section, Config

BASE_URL_PATH = '_ah/api/missinglink/v1/'


def __urljoin(*args):
    base = args[0]
    for u in args[1:]:
        base = parse.urljoin(base, u)

    return base


def __refresh_token_id_token(config, http_session):
    try:
        id_token = update_token(config, http_session)
    except HTTPError:
        eprint('Authorization failed, try running "mali auth init" again')
        sys.exit(1)

    return id_token


def __api_call(config, http_session, http_method_name, url, data):
    id_token = config.id_token if config.id_token else __refresh_token_id_token(config, http_session)

    for _ in range(3):
        headers = {'Authorization': 'Bearer {}'.format(id_token)}
        http_method_exec = getattr(http_session, http_method_name)
        r = http_method_exec(url, headers=headers, json=data)
        if r.status_code == 401:
            id_token = __refresh_token_id_token(config, http_session)
            continue

        r.raise_for_status()
        return r.json()

    eprint('failed to refresh the token, rerun auth init again')
    sys.exit(1)


def add_url_params(url, **params):
    try:
        from urllib import urlencode, unquote
        from urlparse import urlparse, parse_qsl, ParseResult
    except ImportError:
        # Python 3 fallback
        from urllib.parse import (
            urlencode, unquote, urlparse, parse_qsl, ParseResult
        )

    """ Add GET params to provided URL being aware of existing.

    :param url: string of target URL
    :param params: dict containing requested params to be added
    :return: string with updated URL

    >> url = 'http://stackoverflow.com/test?answers=true'
    >> new_params = {'answers': False, 'data': ['some','values']}
    >> add_url_params(url, new_params)
    'http://stackoverflow.com/test?data=some&data=values&answers=false'
    """
    # Unquoting URL first so we don't loose existing args
    url = unquote(url)
    # Extracting url info
    parsed_url = urlparse(url)
    # Extracting URL arguments from parsed URL
    get_args = parsed_url.query
    # Converting URL arguments to dict
    parsed_get_args = dict(parse_qsl(get_args))
    # Merging URL arguments dict with new params
    parsed_get_args.update(params)

    parsed_get_args = collections.OrderedDict(sorted(parsed_get_args.items()))

    # Bool and Dict values should be converted to json-friendly values
    # you may throw this part away if you don't like it :)
    parsed_get_args.update(
        {k: json.dumps(v) for k, v in parsed_get_args.items()
         if isinstance(v, (bool, dict))}
    )

    # Converting URL argument to proper query string
    encoded_get_args = urlencode(parsed_get_args, doseq=True)
    # Creating new parsed result object based on provided with new
    # URL arguments. Same thing happens inside of urlparse.
    new_url = ParseResult(
        parsed_url.scheme, parsed_url.netloc, parsed_url.path,
        parsed_url.params, encoded_get_args, parsed_url.fragment
    ).geturl()

    return new_url


def _handle_async_api(config, http_session, http_method_name, method_url, data, retry):
    if data is not None:
        data['async'] = True
    else:
        method_url = add_url_params(method_url, async=True)

    retry = retry or default_api_retry()
    result = _handle_sync_api(config, http_session, http_method_name, method_url, data, retry)

    if 'token' not in result:
        return result

    token = result['token']

    while True:
        result = _handle_sync_api(config, http_session, 'get', 'data_volumes/tasks/' + token)
        if result.get('failed'):
            raise Exception('Internal Server Error %s' % json.dumps(result))

        if result.get('finished'):
            return json.loads(result['results']) if 'results' in result else None

        time.sleep(2.0)


def _handle_sync_api(config, http_session, http_method_name, method_url, data=None, retry=None):
    import requests

    if config.refresh_token is None:
        eprint('Please run: "mali auth init" to setup authorization')
        sys.exit(1)

    url = __urljoin(config.api_host, BASE_URL_PATH, method_url)

    def api_call_with_retry(*args, **kwargs):
        return retry.call(__api_call, *args, **kwargs)

    f = __api_call if retry is None else api_call_with_retry

    try:
        return f(config, http_session, http_method_name, url, data)
    except requests.exceptions.HTTPError as ex:
        try:
            error_message = ex.response.json().get('error', {}).get('message')
        except ValueError:
            error_message = None

        if error_message is None:
            error_message = str(ex)

        eprint('\n' + error_message)
        sys.exit(1)


def handle_api(ctx_or_config, http_session, http_method_name, method_url, data=None, retry=None, async=False):
    config = ctx_or_config if isinstance(ctx_or_config, Config) else ctx_or_config.config

    if async:
        return _handle_async_api(config, http_session, http_method_name, method_url, data, retry)

    return _handle_sync_api(config, http_session, http_method_name, method_url, data, retry)


def _should_retry_ml_auth(exception):
    import requests

    logging.debug('got retry exception (ml auth) %s', exception)

    error_codes_to_retries = [
        429,  # Too many requests
    ]

    return isinstance(exception, requests.exceptions.HTTPError) and exception.response.status_code in error_codes_to_retries


def update_token(config, http_session):
    from retrying import retry

    @retry(retry_on_exception=_should_retry_ml_auth)
    def with_retry():
        url = __urljoin(config.api_host, BASE_URL_PATH, 'users/refresh_token')

        r = http_session.post(
            url,
            json={
                'refresh_token': config.refresh_token,
            })

        r.raise_for_status()

        data = r.json()

        section = get_prefix_section(config.config_prefix, 'token')
        config.set(section, 'id_token', data['id_token'])
        config.save()

        return data['id_token']

    return with_retry()


def default_api_retry(stop_max_attempt_number=None):
    from retrying import Retrying

    def retry_if_retry_possible_error(exception):
        logging.debug('got retry exception (api) %s', exception)

        return True

    return Retrying(
        retry_on_exception=retry_if_retry_possible_error,
        wait_exponential_multiplier=50,
        wait_exponential_max=5000,
        stop_max_attempt_number=stop_max_attempt_number)



# -*- coding: utf8 -*-
import json
import threading
from contextlib import closing
import requests
import importlib
from .config import Config
from .data_volume import with_repo
from .eprint import eprint
from .exceptions import AccessDenied, NotFound


class DownloadEntity(object):
    local_data = threading.local()

    @classmethod
    def __object_from_data(cls, data, creator):
        data_key = json.dumps(data, sort_keys=True)

        try:
            return cls.local_data.__data_sync_objects[data_key]
        except KeyError:
            cls.local_data.__data_sync_objects[data_key] = creator(data)
        except AttributeError:
            cls.local_data.__data_sync_objects = {data_key: creator(data)}

        return cls.local_data.__data_sync_objects[data_key]

    @classmethod
    def _import_storage(cls, storage_class):
        module_name, class_name = storage_class.rsplit('.', 1)
        m = importlib.import_module(module_name)
        return getattr(m, class_name)

    @classmethod
    def _get_storage(cls, current_data):
        return cls._import_storage(current_data['class']).init_from_config(**current_data)

    @classmethod
    def _get_config(cls, current_data):
        return Config(**current_data)

    @classmethod
    def _get_item_data(cls, repo, storage, metadata):
        if storage.has_item(metadata):
            return

        try:
            _, current_data = repo.object_store.get_raw(metadata)
            return current_data
        except (AccessDenied, NotFound) as ex:
            eprint(ex)
            return
        except requests.exceptions.HTTPError as ex:
            if ex.response.status_code == 404:
                return

            raise

    @classmethod
    def download(cls, config_init_dict, volume_id, metadata, headers):
        config = cls.__object_from_data(config_init_dict, cls._get_config)

        storage_config = config.items('storage')

        storage = cls.__object_from_data(storage_config, cls._get_storage)

        session = requests.session()
        session.headers.update(headers)

        with with_repo(config, volume_id, read_only=True, session=session) as repo:
            with closing(storage):
                data = cls._get_item_data(repo, storage, metadata)
                if data is not None:
                    storage.add_item(metadata, data)



# -*- coding: utf8 -*-
from .auth import auth_commands
from .code import code_commands
from .data import data_commands
from .experiments import experiments_commands
from .models import models_commands
from .orgs import orgs_commands
from .projects import projects_commands
from .run import run_commands
from .resources import resource_commands



# -*- coding: utf8 -*-
from mali_commands.mali_version import get_mali_version


def create_http_session():
    import requests

    session = requests.session()

    session.headers.update({'User-Agent': 'mali/{}'.format(get_mali_version())})

    return session



# -*- coding: utf8 -*-
import collections
import json
import logging
import time

from concurrent.futures import ThreadPoolExecutor

from .download_entity import DownloadEntity
from .metadata_files import MetadataFiles

try:
    # noinspection PyPep8Naming
    import cPickle as pickle
except ImportError:
    import pickle

import random
import string
import tempfile
from contextlib import closing
import sys
import os
from uuid import uuid4
import datetime
from six import wraps
from .api import default_api_retry
from .json_utils import multi_line_json_from_data, normalize_item, dict_normalize
from . import MetadataOperationError
from .multi_process_control import get_multi_process_control
from .path_utils import get_batch_of_files_from_paths, safe_make_dirs, DestPathEnum, enumerate_paths_with_info
from .print_status import PrintStatus
from .eprint import eprint


def wrap_exceptions(fn):
    @wraps(fn)
    def decorated(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except AssertionError:
            raise
        except Exception as ex:
            eprint('\n' + str(ex))
            sys.exit(1)

    return decorated


epoch = datetime.datetime.utcfromtimestamp(0)


class InvalidJsonFile(Exception):
    def __init__(self, filename, ex):
        self.filename = filename
        self.ex = ex


def _download_entity(config_init_dict, volume_id, metadata, headers):
    return DownloadEntity.download(config_init_dict, volume_id, metadata, headers)


def chop_microseconds(delta):
    return delta - datetime.timedelta(microseconds=delta.microseconds)


def status_with_timing(message, callback, fp=sys.stdout, on_summery_data=None):
    start_time = datetime.datetime.utcnow()
    status_line = PrintStatus(fp=fp)
    with closing(status_line):
        status_line.print_status(message)

        with ThreadPoolExecutor(max_workers=1) as executor:
            future = executor.submit(callback)

            while future.running():
                total_time = chop_microseconds(datetime.datetime.utcnow() - start_time)
                status_line.print_status("%s %s" % (message, total_time))

                time.sleep(0.1)

            result = future.result()
            total_time = chop_microseconds(datetime.datetime.utcnow() - start_time)

            if on_summery_data is not None:
                summery_data = on_summery_data(result)
                summery_data = '%s, %s' % (total_time, summery_data)
            else:
                summery_data = str(total_time)

            status_line.print_status('%s Done (%s)' % (message, summery_data))

    return result


def disk_result(resume_token, method, result_callback=None):
    def wrap_f(f):
        def wrap(*args, **kwargs):
            import tempfile

            def loader():
                cache_result = pickle.load(cache_file)
                if result_callback is not None:
                    result_callback(cache_result)

                return cache_result

            name = '{func_name}_{python_version}'.format(func_name=method, python_version=sys.hexversion)
            full_path = os.path.join(tempfile.gettempdir(), 'ml_cache', resume_token, name)

            try:
                with open(full_path, 'rb') as cache_file:
                    return status_with_timing('Loading cache file', loader)
            except IOError:
                pass
            except Exception as ex:
                logging.debug('failed to load %s %s', full_path, ex)

            result = f(*args, **kwargs)

            full_path_tmp = full_path + '.tmp'
            try:
                safe_make_dirs(os.path.dirname(full_path_tmp))
                with open(full_path_tmp, 'wb') as cache_file:
                    pickle.dump(result, cache_file, pickle.HIGHEST_PROTOCOL)
                    os.rename(full_path_tmp, full_path)
            except Exception as ex:
                logging.info('failed to save %s %s', full_path_tmp, ex)

            return result

        return wrap

    return wrap_f


class DiskStorage(object):
    def __init__(self, dest_pattern, save_meta):
        self.__dest_pattern = dest_pattern
        self.__save_meta = save_meta

    def close(self):
        pass

    def add_item(self, metadata, data):
        full_filename = DestPathEnum.get_full_path(self.__dest_pattern, metadata)
        safe_make_dirs(os.path.dirname(full_filename))

        if self.__save_meta:
            item_meta = {key: val for key, val in metadata.items() if not key.startswith('@')}
            with open(full_filename + '.metadata.json', 'w') as f:
                json.dump(item_meta, f)

        with open(full_filename, 'wb') as f:
            f.write(data)

    def has_item(self, metadata):
        full_filename = DestPathEnum.get_full_path(self.__dest_pattern, metadata)
        return os.path.isfile(full_filename)

    @property
    def storage_params(self):
        return {
            'dest_pattern': self.__dest_pattern,
            'save_meta': self.__save_meta,
        }

    # noinspection PyUnusedLocal
    @classmethod
    def init_from_config(cls, dest_pattern, save_meta, **kwargs):
        return cls(dest_pattern, save_meta)


class DataSync(object):
    resume_prefix = 'resume-'

    def __init__(self, ctx, repo, no_progressbar, resume_token=None):
        self.__ctx = ctx
        self.__repo = repo
        self.__no_progressbar = no_progressbar

        self.__resume_token = resume_token or self.generate_resume_token()

        if not self.__resume_token.startswith(self.resume_prefix):
            self.__resume_token = self.resume_prefix + self.__resume_token

    @property
    def repo(self):
        return self.__repo

    @property
    def resume_token(self):
        return self.__resume_token

    def __upload_file_for_processing(self, file_obj, file_description):
        from .gcs_utils import do_upload
        from tqdm import tqdm

        volume_id = self.__repo.data_volume_config.volume_id

        data_object_name = '%s/temp/%s_%s.json' % (volume_id, file_description, uuid4().hex)

        url = 'data_volumes/{volume_id}/gcs_urls'.format(volume_id=volume_id)

        headers = {'Content-Type': 'application/json'}

        msg = {
            'methods': 'PUT',
            'paths': [data_object_name],
            'content_type': 'application/json',
            'temp': True,
        }

        result = self.__ctx.obj.handle_api(self.__ctx.obj, self.__ctx.obj.session, 'post', url, msg, retry=default_api_retry())

        put_url = result['put'][0]

        gcs_auth = self.__repo.data_volume_config.object_store_config.get('auth')

        def update_bar(c):
            bar.update(c)

        logging.debug('uploading %s', data_object_name)
        with tqdm(total=file_obj.tell(), desc="Uploading {}".format(file_description), unit=' KB', ncols=80,
                  disable=self.__no_progressbar, unit_scale=True) as bar:
            do_upload(gcs_auth, None, data_object_name, file_obj, headers, head_url=None, put_url=put_url,
                      callback=update_bar)

        return data_object_name

    def __process_index(self, object_name, isolation_token):
        index = self.__repo.open_index()

        def get_changeset():
            return index.get_changeset('gs://' + object_name, isolation_token)

        change_set = status_with_timing('Server process index', get_changeset)

        total_modify_files = 0
        total_new_files = 0

        files_to_upload = tempfile.TemporaryFile('w+')

        for name, op in change_set:
            if op == 'm':
                total_modify_files += 1
            elif op == 'i':
                total_new_files += 1
            else:
                continue

            data = name + '\n'

            files_to_upload.write(data)

        total_files_to_upload = total_modify_files + total_new_files
        files_to_upload.seek(0)

        return files_to_upload, total_files_to_upload

    def upload_and_update_index(self, index_entities_file, isolation_token):
        object_name = self.__upload_file_for_processing(index_entities_file, 'Index')

        files_to_upload, total_files_to_upload = self.__process_index(object_name, isolation_token)

        return files_to_upload, total_files_to_upload

    def upload_and_update_metadata(self, file_obj, isolation_token=None):
        object_name = self.__upload_file_for_processing(file_obj, 'Metadata')

        def add_data():
            return self.__repo.metadata.add_data_using_url('gs://' + object_name, isolation_token)

        status_with_timing('Server process metadata', add_data)

    def __append_index_file(self, file_info, combined_index_files):
        file_name = file_info['path']

        mtime = file_info['mtime']
        ctime = file_info.get('ctime', mtime)

        mtime = (mtime - epoch).total_seconds()
        ctime = (ctime - epoch).total_seconds()

        if file_info['sys'] == 'local':
            rel_path = os.path.relpath(file_name, self.__repo.data_path)

        else:  # file_info['sys'] == 's3':
            rel_path = file_name

        params = {
            'ctime': ctime,
            'mtime': mtime,
            'size': file_info['size'],
            'sha': file_info['sha'],
            'mode': 0
        }

        if 'url' in file_info:
            params['url'] = file_info['url']

        multi_line_json_from_data({rel_path: params}, combined_index_files, key_name='name')

    def __create_combined_index_and_metadata(self, data_path):
        import humanize
        status_line = PrintStatus()

        def on_result(result):
            data_files_info, metadata_files_list = result

            total_files_size = sum(x['size'] for x in data_files_info.values())

            status_line.print_status(
                'Total files {:,} ({}) (metadata found: {:,})', len(data_files_info),
                humanize.naturalsize(total_files_size), len(metadata_files_list))

        @disk_result(self.resume_token, '__create_combined_index_and_metadata', on_result)
        def with_disk_result():
            total_files = 0
            total_files_size = 0
            total_metadata = 0

            data_files_list = set()
            metadata_files_list = set()
            data_files_info = {}

            with closing(status_line):
                for file_info in enumerate_paths_with_info(data_path):
                    rel_file_name = self.repo.rel_path(file_info['path'])
                    total_files_size += file_info['size']

                    if rel_file_name.endswith(MetadataFiles.metadata_ext):
                        total_metadata += 1
                        metadata_files_list.add(rel_file_name)
                    else:
                        total_files += 1
                        data_files_list.add(rel_file_name)
                        data_files_info[rel_file_name] = file_info

                    status_line.print_status(
                        'Total files {:,} ({}) (metadata found: {:,})', total_files,
                        humanize.naturalsize(total_files_size), total_metadata)

            return data_files_info, metadata_files_list

        return with_disk_result()

    def create_combined_index_and_metadata(self, data_path):
        @disk_result(self.resume_token, 'create_combined_index_and_metadata')
        def with_disk_result():
            data_files_info, metadata_files_list = self.__create_combined_index_and_metadata(data_path)

            files_metadata = MetadataFiles.load_all_metadata(
                self.__repo, data_files_info, metadata_files_list, self.__no_progressbar)

            return files_metadata, data_files_info

        return with_disk_result()

    @wrap_exceptions
    def upload_in_batches(self, files_info, total_files=None, callback=None, isolation_token=None):
        total_files = total_files or len(files_info)
        batch_size = max(min(total_files // 100, 250), 250)  # FIXME: hardcoded

        for files_info_batch in get_batch_of_files_from_paths(files_info, batch_size):
            self.__repo.stage(files_info_batch, callback=callback, isolation_token=isolation_token)

    def download_all(self, query, root_folder, dest_pattern, batch_size, processes, save_meta=True):
        multi_process_control = get_multi_process_control(processes)
        return self.__download_all(multi_process_control, query, root_folder, dest_pattern, batch_size, save_meta=save_meta)

    def __download_normalized_item(self, process_control, storage, normalized_item, callback=None):
        volume_id = self.__repo.data_volume_config.volume_id

        def wrap_callback_func(current_item):
            def wrapper(_):
                if callback is not None:
                    callback(current_item)

            return wrapper

        def fullname(o):
            return o.__module__ + "." + o.__class__.__name__

        config_init_dict = self.__ctx.obj.config.init_dict
        storage_params = storage.storage_params
        storage_params['class'] = fullname(storage)
        config_init_dict['storage'] = storage_params

        return process_control.execute(
            _download_entity,
            args=(config_init_dict, volume_id, normalized_item, self.__ctx.obj.session.headers),
            callback=wrap_callback_func(normalized_item))

    class DownloadIterResults(collections.Iterator):
        def __init__(self, repo, query, batch_size, silent=False):
            self.__query = query

            self.__batch_size = batch_size
            self.__repo = repo
            self.__query_offset = 0
            self.__total_data_points = None
            self.__current_iter = None
            self.__async = batch_size == -1
            self.__silent = silent

            if (self.__batch_size or 0) < 0:
                self.__batch_size = None

        def __get_next_results(self):
            def do_query():
                query_offset = None if not self.__query_offset else self.__query_offset
                return self.__repo.metadata.query(self.__query, async=self.__async, max_results=self.__batch_size, start_index=query_offset)

            if self.__async and not self.__silent:
                def on_summery_data(results):
                    import humanize
                    _, query_current_total_data_points, query_total_size = results

                    return '%s, %s Rows' % (humanize.naturalsize(query_total_size), humanize.intcomma(query_current_total_data_points))

                current_results, current_total_data_points, total_size = status_with_timing('Execute Query', do_query, fp=sys.stderr, on_summery_data=on_summery_data)
            else:
                current_results, current_total_data_points, total_size = do_query()

            def iter_items():
                next_results = current_results
                while True:
                    has_items = False
                    for metadata in next_results:
                        self.__query_offset += 1
                        has_items = True
                        yield normalize_item(metadata)

                    # in async mode we download everything in one pass
                    if self.__async or not has_items:
                        break

                    next_results, _, _ = do_query()

            return iter_items(), current_total_data_points, total_size

        def __len__(self):
            self.__prepare_iter_if_needed()

            return self.__total_data_points

        def __prepare_iter_if_needed(self):
            if self.__current_iter is not None:
                return

            results_iter, self.__total_data_points, total_size = self.__get_next_results()
            self.__current_iter = results_iter

        def __next__(self):
            self.__prepare_iter_if_needed()

            return next(self.__current_iter)

        next = __next__

        def __iter__(self):
            self.__prepare_iter_if_needed()
            return self

    def create_download_iter(self, query, batch_size=None, silent=False):
        return self.DownloadIterResults(self.__repo, query, batch_size, silent)

    def download_items(self, normalized_metadata_items, storage, multi_process_control):
        if len(normalized_metadata_items) == 0:
            return

        futures = []
        for normalized_item in normalized_metadata_items:
            async_result = self.__download_normalized_item(multi_process_control, storage, normalized_item)
            futures.append((async_result, normalized_item))

        for async_result, normalized_item in futures:
            async_result.wait()

    def __download_all(self, multi_process_control, query, root_folder, dest_pattern, batch_size, save_meta=True):
        from tqdm import tqdm

        def handle_item(item):
            file_name = DestPathEnum.get_full_path(dest_pattern, item)
            rel_path = os.path.relpath(file_name, root_folder)
            phase = item.get('phase', 'all')
            phase_meta.setdefault(phase, {})[rel_path] = item

            bar.update()

        download_iter = self.create_download_iter(query, batch_size)

        phase_meta = {}

        def get_next_results():
            try:
                return next(download_iter)
            except StopIteration:
                return None

        normalized_item = get_next_results()

        storage = DiskStorage(dest_pattern, save_meta)

        with tqdm(total=len(download_iter), desc="Downloading", unit=' data point', ncols=80, disable=self.__no_progressbar) as bar:
            with closing(multi_process_control):
                try:
                    while normalized_item is not None:
                        self.__download_normalized_item(multi_process_control, storage, normalized_item, handle_item)

                        normalized_item = get_next_results()
                except MetadataOperationError as ex:
                    eprint(str(ex))
                except KeyboardInterrupt:
                    multi_process_control.terminate()

        return phase_meta

    def save_metadata(self, root_dest, metadata):
        from tqdm import tqdm

        with tqdm(total=len(metadata), desc="saving metadata", unit=' files', ncols=80, disable=self.__no_progressbar) as bar:
            for key, val in metadata.items():
                if key is None:
                    key = 'unknown'

                json_metadata_file = os.path.join(root_dest, key + '.metadata.json')

                with open(json_metadata_file, 'w') as f:
                    json.dump(val, f)

                bar.update()

    def create_index_file_from_data(self, f, files_info):
        for val in files_info.values():
            self.__append_index_file(val, f)

    @classmethod
    def create_metadata_file_from_data(cls, f, metadata_files):
        dict_normalize(metadata_files)

        for rel_file_name, file_metadata in metadata_files.items():
            multi_line_json_from_data({rel_file_name: file_metadata}, f)

    def __get_files_to_upload(self, files_info, combined_index_files, isolation_token):
        files_to_upload, total_files_to_upload = self.upload_and_update_index(combined_index_files, isolation_token)

        def switch_to_full_path(params):
            rel_path = params['path']

            params['path'] = self.repo.full_path(rel_path)

            return params

        def enum_file_names():
            for file_name in files_to_upload:
                file_name = file_name.strip()

                yield file_name

        files_to_upload = [switch_to_full_path(files_info[rel_file_name]) for rel_file_name in enum_file_names()]

        return files_to_upload

    def __upload_full_index(self, index_entities_file, isolation_token):
        object_name = self.__upload_file_for_processing(index_entities_file, 'Index')
        index = self.__repo.open_index()

        def set_entries():
            return index.set_entries('gs://' + object_name, isolation_token)

        status_with_timing('Server process index', set_entries)

    def upload_index_and_metadata(self, data_path, isolation_token=None):
        def with_disk_result():
            metadata_files, files_info = self.create_combined_index_and_metadata(data_path)

            if len(files_info) == 0:
                return []

            if len(metadata_files) > 0:
                combined_meta_files = tempfile.TemporaryFile('wb+')
                self.create_metadata_file_from_data(combined_meta_files, metadata_files)
                self.upload_and_update_metadata(combined_meta_files, isolation_token)

            combined_index_files = tempfile.TemporaryFile('wb+')
            self.create_index_file_from_data(combined_index_files, files_info)

            if self.repo.data_volume_config.embedded:
                return self.__get_files_to_upload(files_info, combined_index_files, isolation_token)

            return self.__upload_full_index(combined_index_files, isolation_token)

        return with_disk_result()

    @classmethod
    def generate_resume_token(cls):
        size = 6
        chars = string.digits
        result = ''.join(random.choice(chars) for _ in range(size))

        return result



# -*- coding: utf8 -*-
import logging
try:
    from .gcs import GCSObjectStore, BackendGCSObjectStore
except ImportError as ex:
    logging.debug("ImportError: %s", ex)
    GCSObjectStore = None
    BackendGCSObjectStore = None

from .null_storage import NullObjectStore



# -*- coding: utf8 -*-
from ..backend_mixin import BackendMixin
from ..dulwich.refs import RefsContainer


SYMREF = b'ref: '


class BackendRefContainer(BackendMixin, RefsContainer):
    def __init__(self, connection, config, session, handle_api):
        super(BackendRefContainer, self).__init__(connection, config, session, handle_api)

    def get_packed_refs(self):
        return {}

    def set_if_equals(self, name, old_ref, new_ref):
        with self._connection.get_cursor() as session:
            url = 'data_volumes/%s/ref/%s' % (self._volume_id, name)

            msg = {
                'old_ref': old_ref,
                'new_ref': new_ref,
            }

            result = self._handle_api(self._config, session, 'patch', url, msg)

            return result.get('ok')

    def add_if_new(self, name, ref):
        with self._connection.get_cursor() as session:
            url = 'data_volumes/%s/ref/%s' % (self._volume_id, name)

            msg = {
                'ref': ref
            }

            result = self._handle_api(self._config, session, 'post', url, msg)

            return result.get('ok')

    def read_loose_ref(self, name):
        with self._connection.get_cursor() as session:
            url = 'data_volumes/%s/ref/%s' % (self._volume_id, name)

            result = self._handle_api(self._config, session, 'get', url)

            commit_sha = result.get('commit_sha')

            if commit_sha is None:
                raise KeyError(name)

    def set_symbolic_ref(self, name, other):
        raise NotImplementedError(self.set_symbolic_ref)

    def allkeys(self):
        raise NotImplementedError(self.allkeys)

    def remove_if_equals(self, name, old_ref):
        raise NotImplementedError(self.remove_if_equals)



# -*- coding: utf8 -*-
import click
from mali_commands.commons import add_to_data_if_not_none, create_validate_length, output_result
from mali_commands.options import project_id_option, alert_type_option


@click.group('projects')
def projects_commands():
    pass


@projects_commands.command('list')
@click.pass_context
def list_projects(ctx):
    projects = ctx.obj.handle_api(ctx.obj, ctx.obj.session, 'get', 'projects')

    output_result(ctx, projects.get('projects', []), ['project_id', 'display_name', 'description', 'token', 'org'])


max_project_display_name = 140
min_project_display_name = 1

max_project_description = 140
min_project_description = 0


@projects_commands.group('alerts')
def alerts_commands():
    pass


@alerts_commands.command('subscribe')
@project_id_option(required=True)
@alert_type_option(required=True, multiple=True)
@click.pass_context
def subscribe_alert(ctx, projectid, alerttype):
    data = {
        'alert_types': alerttype,
        'project_ids': projectid,
    }

    result = ctx.obj.handle_api(ctx.obj, ctx.obj.session, 'post', 'projects/subscribe', data)

    output_result(ctx, result, ['ok'])


@alerts_commands.command('unsubscribe')
@project_id_option(required=True, multiple=True)
@alert_type_option(required=True, multiple=True)
@click.pass_context
def unsubscribe_alert(ctx, projectid, alerttype):
    data = {
        'alert_types': alerttype,
        'project_ids': projectid,
    }

    result = ctx.obj.handle_api(ctx.obj, ctx.obj.session, 'post', 'projects/unsubscribe', data)

    output_result(ctx, result, ['ok'])


@alerts_commands.command('list')
@click.pass_context
def list_alert(ctx):
    alerts = ctx.obj.handle_api(ctx.obj, ctx.obj.session, 'get', 'projects/subscriptions')

    output_result(ctx, alerts.get('alerts', []), ['project_id', 'display_name', 'org', 'alert_type', 'subscribed'])


@projects_commands.command('create')
@click.option(
    '--displayName', required=True, callback=create_validate_length(min_project_display_name, max_project_display_name))
@click.option(
    '--description', required=False, callback=create_validate_length(min_project_description, max_project_description))
@click.option('--org', required=False)
# deprecated option but some clients use it
@click.pass_context
def create_project(ctx, displayname, description, org):
    data = {}

    add_to_data_if_not_none(data, displayname, "display_name")
    add_to_data_if_not_none(data, org, "org")
    add_to_data_if_not_none(data, description, "description")

    result = ctx.obj.handle_api(ctx.obj, ctx.obj.session, 'post', 'projects', data)

    output_result(ctx, result, ['id', 'token'])


@projects_commands.command('transfer')
@project_id_option(required=True)
@click.option('--transferTo', required=False)
@click.pass_context
def transfer_project(ctx, projectid, transferto):
    data = {}

    add_to_data_if_not_none(data, transferto, "transfer_to")

    result = ctx.obj.handle_api(
        ctx.obj, ctx.obj.session, 'post', 'projects/{project_id}/transfer'.format(project_id=projectid), data)

    output_result(ctx, result, ['ok'])



#!/usr/bin/env python
# -*- coding: utf8 -*-
import logging
import sys
import click
import warnings
from mali_commands.http_session import create_http_session
from mali_commands.legit.config import default_missing_link_folder
from mali_commands.legit.context import init_context2
from mali_commands.legit.exceptions import MissingLinkException
from mali_commands.mali_version import get_mali_version, get_mali_package
from self_update.sdk_version import get_keywords
import click_completion
from mali_commands.rsync import makedir

__prev_resolve_ctx = click_completion.resolve_ctx


def __mali_resolve_ctx(_cli, prog_name, args):
    def find_top_parent_ctx(current_ctx):
        parent = current_ctx
        while True:
            if current_ctx.parent is None:
                break

            parent = current_ctx.parent
            current_ctx = parent

        return parent

    ctx = __prev_resolve_ctx(cli, prog_name, args)

    top_ctx = find_top_parent_ctx(ctx)

    init_context2(
        ctx,
        top_ctx.params.get('session'),
        top_ctx.params.get('outputformat'),
        top_ctx.params.get('configprefix'),
        top_ctx.params.get('configfile'))

    return ctx


click_completion.resolve_ctx = __mali_resolve_ctx
click_completion.init()

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'], color=True)


def print_banner():
    from pyfiglet import figlet_format
    from termcolor import cprint
    from colorama import init

    init(strip=not sys.stdout.isatty())  # strip colors if stdout is redirected

    current_version = get_mali_version()

    cprint(figlet_format('MALI   %s' % current_version, font='big'), 'yellow')


def _setup_logger(log_level):
    if not log_level:
        return

    log_level = log_level.upper()

    root_logger = logging.getLogger()
    root_logger.setLevel(log_level)
    handler = logging.StreamHandler(stream=sys.stdout)
    formatter = logging.Formatter('%(asctime)s.%(msecs)03d %(name)s %(levelname)s %(message)s', datefmt='%Y-%m-%dT%H:%M:%S')
    handler.setFormatter(formatter)
    root_logger.addHandler(handler)

    logging_method = getattr(root_logger, log_level.lower())
    logging_method('log level set to %s (this is a test message)', log_level)


class Choice2(click.ParamType):
    """The choice type allows a value to be checked against a fixed set of
    supported values.  All of these values have to be strings.
    :param case_sensitive: Set to false to make choices case insensitive.
    Defaults to true.
    See :ref:`choice-opts` for an example.
    """
    name = 'choice2'

    def __init__(self, choices, case_sensitive=False):
        self.choices = choices
        self.case_sensitive = case_sensitive

    def get_metavar(self, param):
        return '[%s]' % '|'.join(self.choices)

    def get_missing_message(self, param):
        return 'Choose from %s.' % ', '.join(self.choices)

    def _normalize_value(self, ctx, value):
        # Match through normalization and case sensitivity
        # first do token_normalize_func, then lowercase
        # preserve original `value` to produce an accurate message in
        # `self.fail`
        if ctx is not None and ctx.token_normalize_func is not None:
            normed_value = ctx.token_normalize_func(value)
            normed_choices = [ctx.token_normalize_func(choice) for choice in self.choices]
        else:
            normed_value = value
            normed_choices = self.choices

        return normed_value, normed_choices

    def _get_normed_choices_value(self, ctx, value):
        normed_value, normed_choices = self._normalize_value(ctx, value)

        if not self.case_sensitive:
            normed_value = normed_value.lower()
            normed_choices = [choice.lower() for choice in normed_choices]

        return normed_value, normed_choices

    def convert(self, value, param, ctx):
        # Exact match
        if value in self.choices:
            return value

        normed_value, normed_choices = self._get_normed_choices_value(ctx, value)

        if normed_value in normed_choices:
            return normed_value

        self.fail('invalid choice: %s. (choose from %s)' % (value, ', '.join(self.choices)), param, ctx)

    def __repr__(self):
        return 'Choice(%r)' % list(self.choices)


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option('--outputFormat', '-o', type=Choice2(['tables', 'json', 'csv']), default='tables', required=False)
@click.option('--configPrefix', '-cp', envvar='ML_CONFIG_PREFIX', required=False)
@click.option('--configFile', '-cf', envvar='ML_CONFIG_FILE', required=False)
@click.option('--logLevel', envvar='ML_LOG_LEVEL', type=Choice2(['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG']), required=False)
@click.pass_context
def cli(ctx, outputformat, configprefix, configfile, loglevel):
    from tqdm import TqdmSynchronisationWarning

    warnings.filterwarnings('ignore', category=TqdmSynchronisationWarning)

    _setup_logger(loglevel)
    init_context2(ctx, create_http_session(), outputformat, configprefix, configfile)


@cli.command()
@click.argument('shell', required=False, type=click_completion.DocumentedChoice(click_completion.shells))
def install(shell):
    """Install the click-completion-command completion"""
    from mali_commands.install import rc_updater

    shell = shell or click_completion.get_auto_shell()

    code = click_completion.get_code(shell)

    file_name = '{dir}/completion.{shell}.inc'.format(dir=default_missing_link_folder(), shell=shell)

    makedir(file_name)

    with open(file_name, 'w') as f:
        f.write(code)

    rc_updater(shell, file_name)


@cli.command('version')
@click.pass_context
def version(_ctx):
    current_version = get_mali_version()

    print_banner()

    click.echo('%s %s' % (get_mali_package(), current_version))


# noinspection PyBroadException
def update_sdk(latest_version, user_path, throw_exception):
    from self_update.pip_util import pip_install, get_pip_server

    keywords = get_keywords(get_mali_package()) or []

    require_package = '%s==%s' % (get_mali_package(), latest_version)
    p, args = pip_install(get_pip_server(keywords), require_package, user_path)

    if p is None:
        return False

    try:
        std_output, std_err = p.communicate()
    except Exception:
        if throw_exception:
            raise

        logging.exception("%s failed", " ".join(args))
        return False

    rc = p.returncode

    if rc != 0:
        logging.error('%s failed to upgrade to latest version (%s)', get_mali_package(), latest_version)
        logging.error("failed to run %s (%s)\n%s\n%s", " ".join(args), rc, std_err, std_output)
        return False

    logging.info('%s updated to latest version (%s)', get_mali_package(), latest_version)

    return True


def self_update(throw_exception=False):
    from self_update.pip_util import get_latest_pip_version

    current_version = get_mali_version()
    keywords = get_keywords(get_mali_package()) or []

    if current_version is None:
        return

    latest_version = get_latest_pip_version(get_mali_package(), keywords, throw_exception=throw_exception)

    if latest_version is None:
        return

    if current_version == latest_version:
        return

    running_under_virtualenv = getattr(sys, 'real_prefix', None) is not None

    if not running_under_virtualenv:
        logging.info('updating %s to version %s in user path', get_mali_package(), latest_version)

    return update_sdk(latest_version, user_path=not running_under_virtualenv, throw_exception=throw_exception)


def add_commands():
    from mali_commands import auth_commands, projects_commands, orgs_commands, experiments_commands, code_commands, \
        models_commands, data_commands, run_commands, resource_commands

    cli.add_command(auth_commands)
    cli.add_command(projects_commands)
    cli.add_command(orgs_commands)
    cli.add_command(experiments_commands)
    cli.add_command(code_commands)
    cli.add_command(models_commands)
    cli.add_command(data_commands)
    cli.add_command(run_commands)
    cli.add_command(resource_commands)


def main():
    import os

    if os.environ.get('MISSINGLINKAI_ENABLE_SELF_UPDATE'):
        self_update()

    add_commands()
    cli()


if __name__ == "__main__":
    try:
        main()
    except MissingLinkException as ex:
        click.echo(ex, err=True)



# -*- coding: utf8 -*-
import datetime
import logging
from ..bigquery_mixin import BigQueryMixin, BqJob
from .base_db_index import BaseMLIndex


# noinspection SqlNoDataSourceInspection,SqlResolve
class BigQueryMLIndex(BaseMLIndex, BigQueryMixin):
    STAGING_INDEX_TABLE_NAME = 'staging_index'
    INDEX_TABLE_NAME = 'index'

    def __init__(self, connection, version=None, delete_temp_on_commit=True, isolated=False):
        self.__version = version or 0

        super(BigQueryMLIndex, self).__init__(connection)
        self.__delete_temp_on_commit = delete_temp_on_commit
        self.__isolated = isolated

    def __get_index_table_ref(self):
        return self._get_table_ref(self.INDEX_TABLE_NAME)

    def __get_staging_index_table_ref(self, version=None):
        staging_table_name = '%s_%s' % (self.STAGING_INDEX_TABLE_NAME, version or self.__version or 0)
        return self._get_table_ref(staging_table_name)

    @classmethod
    def __table_schema(cls):
        from google.cloud import bigquery

        schema = (
            bigquery.SchemaField('name', 'STRING', 'REQUIRED'),
            bigquery.SchemaField('sha', 'STRING', 'REQUIRED'),
            bigquery.SchemaField('ctime', 'FLOAT', 'REQUIRED'),
            bigquery.SchemaField('mtime', 'FLOAT', 'REQUIRED'),
            bigquery.SchemaField('mode', 'INTEGER', 'REQUIRED'),
            bigquery.SchemaField('size', 'INTEGER', 'REQUIRED'),
            bigquery.SchemaField('url', 'STRING'),
            bigquery.SchemaField('commit_sha', 'STRING', 'REQUIRED'),
            bigquery.SchemaField('ts', 'TIMESTAMP', 'REQUIRED'),  # this has to be the last column
        )

        return schema

    def __create_table_by_ref_if_needed(self, table_ref):
        logging.info('create table if needed %s', table_ref.table_id)

        import google.cloud.exceptions
        from google.cloud.bigquery.table import Table

        bq_client = self._connection

        table = Table(table_ref)
        table.schema = self.__table_schema()

        try:
            bq_client.create_table(table)
        except google.cloud.exceptions.Conflict:
            pass

        return table

    def _create_table_if_needed(self):
        self.__create_table_by_ref_if_needed(self.__get_index_table_ref())
        self.__create_table_by_ref_if_needed(self.__get_staging_index_table_ref())

    def set_entries(self, entries, dry_mode=False):
        if not entries:
            return

        now = datetime.datetime.utcnow()

        def decode_row_without_gid_uid(entries):
            for name, sha, ctime, mtime, mode, _, _, size, url in self._decode_entries(entries):
                yield name, sha, ctime, mtime, mode, size, url, 'staging', now

        rows = [row for row in decode_row_without_gid_uid(entries)]

        logging.debug('inserting %s rows into bq', len(rows))

        staging_index_table_ref = self.__get_staging_index_table_ref()

        bq_client = self._connection

        if not dry_mode:
            bq_client.create_rows(staging_index_table_ref, rows, selected_fields=self.__table_schema())

        logging.debug('inserted %s rows into bq', len(rows))

    def __truncate_staging_index(self):
        if not self.__delete_temp_on_commit:
            logging.debug('index: delete_temp_on_commit: False')
            return

        logging.info('truncate index staging')

        staging_table_ref = self.__get_staging_index_table_ref()
        bq_client = self._connection
        bq_client.delete_table(staging_table_ref)

    def __query_into_index_table(self, from_table_ref, to_table_ref, commit_sha, ts, remove_duplicates=True):
        from google.cloud import bigquery

        bq_client = self._connection

        with bq_client.get_cursor() as bq_dataset:
            if remove_duplicates:
                src_query = """
                    #standardSQL
                    #__query_into_index_table
                    SELECT interim_index_table.*, @commit_sha as commit_sha, @ts as ts 
                    FROM (
                        SELECT * EXCEPT(row_number, ts, commit_sha)
                        FROM (
                          SELECT *, ROW_NUMBER() OVER (PARTITION BY name ORDER BY ts DESC) row_number
                          FROM `{dataset_name}.{from_table_ref}`
                        )
                        WHERE row_number = 1
                    ) interim_index_table
                  """.format(
                    dataset_name=bq_dataset.dataset_id,
                    from_table_ref=from_table_ref.table_id,
                )
            else:
                src_query = """
                    #standardSQL
                    #__query_into_index_table
                    SELECT interim_index_table.*, @commit_sha as commit_sha, @ts as ts 
                    FROM `{dataset_name}.{from_table_ref}` interim_index_table
                  """.format(
                    dataset_name=bq_dataset.dataset_id,
                    from_table_ref=from_table_ref.table_id,
                )

            src_query_parameters = (
                bigquery.ScalarQueryParameter('commit_sha', 'STRING', commit_sha),
                bigquery.ScalarQueryParameter('ts', 'TIMESTAMP', ts),
            )

            job = self._async_copy_table_data(
                src_query, src_query_parameters, to_table_ref)

            return BqJob(job)

    def begin_commit(self, commit_sha, tree_id, ts):
        staging_index_table_ref = self.__get_staging_index_table_ref()

        if not self.__isolated:
            self.__create_table_by_ref_if_needed(self.__get_staging_index_table_ref(version=self.__version + 1))

        return self.__query_into_index_table(staging_index_table_ref, self.__get_index_table_ref(), commit_sha, ts)

    def end_commit(self):
        if not self.__isolated:
            self.__truncate_staging_index()

    def delete_all(self):
        index_table = self._get_table_name(self._connection.table_prefix, self.INDEX_TABLE_NAME)
        staging_index_table_prefix = self._get_table_name(self._connection.table_prefix, self.STAGING_INDEX_TABLE_NAME)
        self._connection.delete_tables([index_table, staging_index_table_prefix])

    def _query_changeset(self, new_index_table):
        staging_index_table_ref = self.__get_staging_index_table_ref()
        index_table_ref = self.__get_index_table_ref()

        with self._connection.get_cursor() as bq_dataset:
            query = """
                #standardSQL
                #_query_changeset
                SELECT name, op
                FROM (
                    SELECT TableA.name as name, 'c' as op
                    FROM `{dataset_name}.{new_index_table}` AS TableA
                    LEFT JOIN (
                      SELECT sha, name
                      FROM `{dataset_name}.{staging_index_table}`
                      UNION ALL
                      SELECT sha, name
                      FROM `{dataset_name}.{index_table}`
                    ) AS TableB
                    ON TableA.name = TableB.name 
                    WHERE TableA.sha != TableB.sha
                    UNION ALL
                    SELECT TableA.name as name, 'i' as op
                    FROM `{dataset_name}.{new_index_table}` AS TableA
                    LEFT JOIN (
                      SELECT sha, name
                      FROM `{dataset_name}.{staging_index_table}`
                      UNION ALL
                      SELECT sha, name
                      FROM `{dataset_name}.{index_table}`
                    ) AS TableB
                    ON TableA.name = TableB.name AND TableA.sha = TableB.sha
                    WHERE TableB.sha IS NULL
                    UNION ALL
                    SELECT TableA.name as name, 'd' as op
                    FROM `{dataset_name}.{new_index_table}` AS TableA
                    RIGHT OUTER JOIN (
                      SELECT sha, name
                      FROM `{dataset_name}.{staging_index_table}`
                      UNION ALL
                      SELECT sha, name
                      FROM `{dataset_name}.{index_table}`
                    ) AS TableB
                    ON TableA.name = TableB.name
                    WHERE TableA.name is NULL
                )        
            """.format(
                dataset_name=bq_dataset.dataset_id,
                new_index_table=new_index_table.table_id,
                staging_index_table=staging_index_table_ref.table_id,
                index_table=index_table_ref.table_id,
            )

        return self._query_async(query)

    def _load_table_from_url(self, table_url, table_ref=None):
        job = self._async_load_job(table_url, table_ref=table_ref)

        result = BqJob(job).wait()

        bq_client = self._connection

        return bq_client.get_table(result.destination)

    def add_data_using_url(self, index_url, changeset_result_url):
        data_interim_table = self._load_table_from_url(index_url)

        bq_client = self._connection

        if changeset_result_url is not None:
            query_job = self._query_changeset(data_interim_table)

            BqJob(query_job).wait()

            changeset_table = bq_client.get_table(query_job.destination)

            extract_job = self._extract_table(changeset_table, changeset_result_url)

            BqJob(extract_job).wait()
        else:
            now = datetime.datetime.utcnow()
            staging_index_table_ref = self.__get_staging_index_table_ref()
            self.__query_into_index_table(data_interim_table, staging_index_table_ref, 'staging', now, remove_duplicates=False).wait()

        bq_client.delete_table(data_interim_table)

    def delete_version(self, version_id):
        from google.cloud import bigquery

        query_parameters = (
            bigquery.ScalarQueryParameter('version_id', 'STRING', version_id),
        )

        index_table_ref = self.__get_index_table_ref()
        with self._connection.get_cursor() as bq_dataset:
            query = '''
                #standardSQL
                #delete_version
                SELECT * 
                FROM `{dataset_name}.{index_table_name}`
                WHERE `commit_sha` <> @version_id 
            '''.format(
                dataset_name=bq_dataset.dataset_id,
                index_table_name=index_table_ref.table_id,
            )

            return self.override_table(query, query_parameters, index_table_ref)



import boto3
from botocore.exceptions import NoCredentialsError, ClientError
# from botocore.errorfactory import
import logging
import os
import sys
import json

FORMAT = '%(name) %(asctime)-15s  %(levelname)-8s %(message)s'
logging.basicConfig(level=logging.INFO)
logging.getLogger('botocore.credentials').setLevel(logging.WARNING)
logging.getLogger('botocore.vendored.requests.packages.urllib3.connectionpool').setLevel(logging.WARNING)
logger = logging.getLogger(__file__)


class AwsAuthenticationHelper:
    ML_ACCOUNT_ID = '502812207255'
    ML_ROLE_NAME = 'MissingLinkResourceManager'
    ML_ROLE_DESCRIPTION = 'This policy is used in order to grant permissions to MissingLink.ai to manage the resource managers in this aws account'

    @classmethod
    def _get_script_path(cls):
        return os.path.dirname(os.path.realpath(__file__))

    @classmethod
    def get_policy(cls):
        with open(os.path.join(cls._get_script_path(), 'aws_policy.json'), 'r') as f:
            return json.load(f)

    @classmethod
    def get_assume_role_policy(cls):
        with open(os.path.join(cls._get_script_path(), 'aws_sts_assume_policy.json'), 'r') as f:
            return json.load(f)

    @classmethod
    def _set_policy(cls):
        target_arn = 'arn:aws:iam::{}:policy/{}'.format(cls.account_id(), cls.ML_ROLE_NAME)
        try:
            policy = cls.iam().get_policy(PolicyArn=target_arn)
        except cls.iam().exceptions.NoSuchEntityException:
            policy = None

        if policy is None:
            logging.info('Creating policy... ')
            policy = cls.iam().create_policy(
                PolicyName=cls.ML_ROLE_NAME,
                Description=cls.ML_ROLE_DESCRIPTION,
                PolicyDocument=json.dumps(cls.get_policy())
            )

        versions = cls.iam().list_policy_versions(PolicyArn=target_arn)
        for version in versions['Versions']:
            if not version['IsDefaultVersion']:
                logging.info('%s: Removing depricated, non default version: %s', cls.ML_ROLE_NAME, version['VersionId'])
                cls.iam().delete_policy_version(PolicyArn=target_arn, VersionId=version['VersionId'])

        cls.iam().create_policy_version(
            PolicyArn=policy['Policy']['Arn'],
            PolicyDocument=json.dumps(cls.get_policy()),
            SetAsDefault=True
        )
        logger.info('New policy version created')
        return policy

    @classmethod
    def _set_attach_policy(cls, role_policy_arn):
        try:
            role = cls.iam().get_role(RoleName=cls.ML_ROLE_NAME)

        except cls.iam().exceptions.NoSuchEntityException:
            logger.info('Creating role %s', cls.ML_ROLE_NAME)
            role = cls.iam().create_role(
                RoleName=cls.ML_ROLE_NAME,
                Description=cls.ML_ROLE_DESCRIPTION,
                AssumeRolePolicyDocument=json.dumps(cls.get_assume_role_policy()),
            )
        attached_policies = cls.iam().list_role_policies(RoleName=cls.ML_ROLE_NAME)['PolicyNames']

        if cls.ML_ROLE_NAME not in attached_policies:
            logging.info("Attaching policy %s to role %s", role_policy_arn, cls.ML_ROLE_NAME)
            cls.iam().attach_role_policy(PolicyArn=role_policy_arn, RoleName=cls.ML_ROLE_NAME)
        return role

    @classmethod
    def grant_access(cls):
        try:
            policy = cls._set_policy()
            policy_arn = policy['Policy']['Arn']
            res = cls._set_attach_policy(policy_arn)
            return res['Role']['RoleId'], res['Role']['Arn']
        except ClientError as ex:
            if ex.response['Error']['Code'] == 'AccessDenied':
                return None

            raise

    @classmethod
    def client(cls, client_id, aws_access_key_id=None, aws_secret_access_key=None, aws_session_token=None):
        return boto3.client(
            client_id,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token)

    @classmethod
    def sts(cls):
        return cls.client('sts')

    @classmethod
    def iam(cls):
        return cls.client('iam')

    @classmethod
    def whoami(cls):
        return cls.sts().get_caller_identity()

    @classmethod
    def account_id(cls):
        return cls.whoami()['Account']

    @classmethod
    def account_name(cls):
        try:
            aliases = cls.iam().list_account_aliases().get('AccountAliases', [])
            if not aliases:
                return cls.account_id()

            return aliases[0]
        except ClientError:
            return cls.account_id()

    @classmethod
    def is_aws_configured(cls):
        try:
            cls.whoami()
            return True
        except NoCredentialsError:
            return False



# -*- coding: utf8 -*-


class MissingLinkException(Exception):
    pass


class NonRetryException(MissingLinkException):
    pass


class AccessDenied(NonRetryException):
    pass


class NotFound(NonRetryException):
    pass



# -*- coding: utf8 -*-
import os
import shutil
import errno
import tempfile as tmp
from contextlib import contextmanager


def safe_make_dirs(dir):
    try:
        os.makedirs(dir)
    except OSError as ex:
        if ex.errno != errno.EEXIST:
            raise


def safe_rm_tree(path):
    try:
        shutil.rmtree(path)
    except OSError as ex:
        if ex.errno != errno.ENOENT:
            raise


@contextmanager
def tempfile(suffix='', dir_name=None):
    def remove():
        try:
            os.remove(tf.name)
        except OSError as e:
            if e.errno != 2:
                raise

    tf = tmp.NamedTemporaryFile(delete=False, suffix=suffix, dir=dir_name)
    tf.file.close()
    try:
        yield tf.name
    finally:
        remove()


@contextmanager
def open_atomic(file_path, *args, **kwargs):
    fsync = kwargs.get('fsync', False)

    with tempfile(dir_name=os.path.dirname(os.path.abspath(file_path))) as tmp_path:
        with open(tmp_path, *args, **kwargs) as file_:
            try:
                yield file_
            finally:
                if fsync:
                    file_.flush()
                    os.fsync(file_.fileno())

        os.rename(tmp_path, file_path)



# -*- coding: utf8 -*-
import io
import json

import logging
import os
import sys
from contextlib import contextmanager

import requests
import six

from ..scam import QueryParser
from ..api import default_api_retry
from ..backend_mixin import BackendMixin
from .base_metadata_db import BaseMetadataDB, MetadataOperationError
from six.moves.urllib.parse import urlencode


class WrapUnicodePy2(object):
    def __init__(self, fp):
        self.__fp = fp

    def write(self, s):
        return self.__fp.write(s)

    def seek_to_beginning(self):
        return self.__fp.seek(0)


class WrapUnicodePy3(object):
    def __init__(self, fp):
        self.__fp = fp

    def write(self, s):
        return self.__fp.write(s.encode('utf8'))

    def seek_to_beginning(self):
        return self.__fp.seek(0)


class _StreamArray(list):
    def __init__(self, total_rows):
        super(_StreamArray, self).__init__()
        self._stream = None
        self._total_rows = total_rows

    def __iter__(self):
        return self._csv_enum(self._stream)

    # according to the comment below
    def __len__(self):
        return self._total_rows

    @classmethod
    def _csv_enum(cls, stream):
        import csv

        reader = csv.DictReader(stream)
        replace_names = {'_sha': 'path', '_hash': 'id', '_commit_sha': 'version'}
        for row in reader:
            row_params = {}
            meta = {}
            for key, val in row.items():
                if val == '':
                    continue

                is_meta = key.startswith('_')
                if is_meta:
                    key = replace_names.get(key, key[1:])
                    row_params[key] = val
                    continue

                meta[key] = val

            if meta:
                row_params['meta'] = meta

            yield row_params


class _StreamArrayFromUrl(_StreamArray):
    def __init__(self, total_rows, result_url):
        super(_StreamArrayFromUrl, self).__init__(total_rows)
        r = requests.get(result_url, stream=True)  # allowed to use requests

        if r.encoding is None:
            r.encoding = 'utf-8'

        self._stream = r.iter_lines(decode_unicode=True)


class _CachedSessionWrapper(object):
    def __init__(self, session, cache_folder_full_path):
        from requests_cache import CachedSession

        self.__session = session
        self._first_request = None
        self.__cached_session = CachedSession(cache_folder_full_path)  # used as utility to store requests
        self.__wrap_cached_session_send(self.__session)

    @classmethod
    def convert_result_to_response(cls, result_url, total_rows, total_size, explicit_query):
        response_data = {
            'ok': True,
            'total_data_points': total_rows,
            'total_size': total_size,
            'explicit_query': explicit_query,
            'data_points': _StreamArrayFromUrl(total_rows, result_url)
        }

        raw_bytes = io.BytesIO()
        wrap_unicode_class = WrapUnicodePy2 if six.PY2 else WrapUnicodePy3
        raw_string = wrap_unicode_class(raw_bytes)
        json.dump(response_data, raw_string)
        raw_string.seek_to_beginning()

        response = requests.Response()
        response.raw = raw_bytes
        response.status_code = 200

        return response

    def store_response(self, response):
        response.request = self._first_request
        cache_key = self.__cached_session.cache.create_key(self._first_request)
        self.__cached_session.cache.save_response(cache_key, response)

    # the first request is the "real" requests (the next are async response check)
    # we store only the first request, check if there is a cached response attached to it
    # in case there is a cached response we send using the cached session class in case it needs to further
    # handle the cache entry
    def __handle_first_request(self, request):
        if self._first_request is not None:
            return None

        self._first_request = request

        cache_key = self.__cached_session.cache.create_key(self._first_request)

        response, timestamp = self.__cached_session.cache.get_response_and_time(cache_key)

        return response

    def __wrap_cached_session_send(self, session):
        prev_session_send = session.send

        def wrapped_send(request, **kwargs):
            session.send = prev_session_send
            response = self.__handle_first_request(request)
            if response is not None:
                return response

            return prev_session_send(request, **kwargs)

        session.send = wrapped_send


class BackendMetadataDB(BackendMixin, BaseMetadataDB):
    max_query_retry = 3
    default_cache_file_name = 'missinglink_query_v' + str(sys.version_info[0])

    def __init__(self, connection, config, session, handle_api, cache_folder=None):
        super(BackendMetadataDB, self).__init__(connection, config, session, handle_api)
        self.__query_parser = QueryParser()

        if cache_folder:
            self.__cache_folder_full_path = os.path.join(cache_folder, self.default_cache_file_name)
        else:
            self.__cache_folder_full_path = self.default_cache_file_name

    def _create_table(self):
        pass

    def _query_head_data(self, sha_list):
        with self._connection.get_cursor() as session:
            url = 'data_volumes/%s/metadata/head' % self._volume_id
            msg = {
                'sha': sha_list,
            }

            result = self._handle_api(
                self._config, session, 'post', url, msg,
                retry=default_api_retry(stop_max_attempt_number=self.max_query_retry))

            for data_item in result.get('metadata_json') or []:
                yield json.loads(data_item)

    def _add_missing_columns(self, data_object):
        pass

    def get_data_for_commit(self, sha, commit_sha):
        raise NotImplementedError(self.get_data_for_commit)

    def _add_data(self, data):
        pass

    def add_data_using_url(self, metadata_url, isolation_token):
        if not metadata_url:
            logging.debug('no data provided')
            return

        logging.debug('add data %s', metadata_url)

        with self._connection.get_cursor() as session:
            url = 'data_volumes/%s/metadata/head/add' % self._volume_id

            msg = {
                'metadata_url': metadata_url,
                'isolation_token': isolation_token,
            }

            return self._handle_api(self._config, session, 'post', url, msg, retry=default_api_retry(), async=True)

    @classmethod
    def _create_iter_data(cls, result):
        def handle_meta_item(val, data_point, result_data_point):
            if isinstance(val, dict):
                result_data_point['meta'] = val
                return

            meta = result_data_point.setdefault('meta', {})
            for meta_key_val in data_point['meta']:
                meta[meta_key_val['key']] = meta_key_val.get('val')

        def _iter_data():
            for data_point in result.get('data_points') or []:
                result_data_point = {}

                for key, val in data_point.items():
                    if key == 'meta':
                        handle_meta_item(val, data_point, result_data_point)
                    else:
                        result_data_point[key] = val

                yield result_data_point

        return _iter_data()

    def __is_stable_query(self, query_text):
        from ..scam import visit_query
        from .version_visitor import VersionVisitor

        tree = self.__query_parser.parse_query(query_text)

        version_visitor = VersionVisitor()
        visit_query(version_visitor, tree)

        is_stable_version = version_visitor.version not in ['head', 'staging']

        return is_stable_version

    def query(self, query_text, **kwargs):
        version_query = query_text if query_text else '@version:head'

        async = kwargs.pop('async', False)

        is_stable_version = self.__is_stable_query(query_text)

        if is_stable_version:
            session_wrapper = _CachedSessionWrapper(self._session, self.__cache_folder_full_path)
        else:
            logging.info('not a stable query, caching cannot be used')
            session_wrapper = None

        params = {
            'query': version_query
        }

        for key, val in kwargs.items():
            if val is None:
                continue

            params[key] = val

        url = 'data_volumes/%s/query/?%s' % (self._volume_id, urlencode(params))

        result = self._handle_api(
            self._config, self._session, 'get', url, retry=default_api_retry(stop_max_attempt_number=self.max_query_retry), async=async)

        is_live_async_response = async and isinstance(result, (list, tuple))
        if is_live_async_response:
            result_url, total_rows, total_size, explicit_query = result

            response = _CachedSessionWrapper.convert_result_to_response(result_url, total_rows, total_size, explicit_query)

            if session_wrapper is not None:
                session_wrapper.store_response(response)

            result = response.json()
        elif not result['ok']:
            raise MetadataOperationError(result['error'])

        return self._create_iter_data(result), int(result.get('total_data_points', 0)), int(result.get('total_size', 0))

    def _query(self, sql_vars, select_fields, where, **kwargs):
        raise NotImplementedError(self._query)

    def get_all_data(self, sha):
        raise NotImplementedError(self.get_all_data)

    def end_commit(self):
        raise NotImplementedError(self.end_commit)

    def begin_commit(self, commit_sha, tree_id, ts):
        raise NotImplementedError(self.begin_commit)



# -*- coding: utf8 -*-
import os

import setuptools
from setuptools import setup, find_packages
import setuptools.command.bdist_egg

build = os.environ.get('PIP_BUILD', 100)
keywords = os.environ.get('PIP_KEYWORDS', 'test')

version = '0.{}'.format(build)


def read_requirements(name=None):
    name = 'requirements.txt' if name is None else '{}-requirements.txt'.format(name)
    with open(name) as f:
        return [r for r in f.readlines() if len(r.strip()) > 0 and r.strip()[0] != '#']


setup(
    name='mali',
    version=version,
    description='Command line tool for missinglink.ai platform',
    author='missinglink.ai',
    author_email='support+mali@missinglink.ai',
    url='https://missinglink.ai',
    license='Apache',
    py_modules=['mali'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 2.7'
    ],
    packages=find_packages(),
    install_requires=read_requirements(),
    extras_require={
        'docker': read_requirements('docker'),
    },
    package_data={
        # If any package contains *.txt or *.rst files, include them:
        'mali_commands.utilities': ['*.json'],
    },
    entry_points='''
        [console_scripts]
        mali=mali:main
    ''',
)



# -*- coding: utf8 -*-
from .bigquery_metadata_db import BigQueryMetadataDB
from .backend_metadata import BackendMetadataDB
from .base_metadata_db import BaseMetadataDB, MetadataTypeNotSupported, MetadataOperationError



# -*- coding: utf8 -*-
import click
from .commons import output_result
from .utilities import source_tracking


@click.group('code', help='Repository tracking allows you to quickly see, compare and clone snapshots, including'
                          ' unpushed, uncommited and unstaged files of your source code for every experiment.'
                          ' In order to use this feature, you must first create setup tracking repository - empty git repository'
                          ' in the same organization as your tracked (source) repository, and then call "track"'
                          ' This feature relays on the same git authentication mechanism as your source repository'
                          ' So please make sure that the ssh key used for fetching your git code has fetch/push permissions'
                          ' on the tracking repository')
def code_commands():
    pass


def _validate_repos_validity(source_repository, tracking_repository):
    if source_tracking.remote_url_obj(source_repository) is None:
        raise click.BadOptionUsage("Source repository: {} is invalid".format(source_repository))

    if source_tracking.remote_url_obj(tracking_repository) is None:
        raise click.BadOptionUsage("Tracking repository: {} is invalid".format(tracking_repository))

    if source_tracking.validate_diff_urls(source_repository, tracking_repository):
        raise click.BadOptionUsage("The repository repository can't be the same as the source one")


def _validate_code_params(source_repository, tracking_repository):
    _validate_repos_validity(source_repository, tracking_repository)

    if not source_tracking.validate_remote_urls_org(source_repository, tracking_repository):
        raise click.BadOptionUsage("The tracking repository must belong to the same organization as the source ({})".format(source_tracking.get_org_from_remote_url(source_repository)))


def _get_source_tracking_remote_url(ctx, param, value):
    if value is None:
        value = source_tracking.get_remote_url()

    if value is None:
        raise click.MissingParameter('--source-repository')

    return value


@code_commands.command('track', help='Sets up tracking for given source repository. When setting up tracking from mali, both source and target repositories must sit under the same organization. For more information please refer to our documentation.')
@click.option('--org', required=True, help="Organization/Workspace. Repository tracking is available to all users that have permission to access the Organization/Workspace")
@click.option('--source-repository', required=False, callback=_get_source_tracking_remote_url, help='Source repository url. Defaults to the first remote of the repository of the current path')
@click.option('--tracking-repository', required=True,
              help='Target repository. When setting up tracking using mali, this repository must be under the same organization as teh source repository (i.e. if the source repo is https://github.com/my_org/my_repo, the tracking repository must be in the https://github.com/my_org organization')
@click.pass_context
def track_add(ctx, org, source_repository, tracking_repository):
    _validate_code_params(source_repository, tracking_repository)

    data = {
        'target': tracking_repository
    }
    target_url = '{org}/code/track?src={src_repo}'.format(org=org, src_repo=source_tracking.remote_to_template(source_repository))
    result = ctx.obj.handle_api(ctx.obj, ctx.obj.session, 'get', target_url, data)
    output_result(ctx, result, ['ok'])
